/*
 *  $Id: adaptation.h,v 1.24 2000/11/02 16:48:24 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: This headerfile defines the interface to the adaption-module.
 *          The adaption-module encapsulates the socket-interface.
 *          The adaption-module also handles timers in the SCTP and its ULP.
 *          .....
 */

/* ############################################################################## */
/* INCLUDES                                                                       */
/* ############################################################################## */

#ifndef ADAPTATION_H
#define ADAPTATION_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "distribution.h"


int adl_str2sockunion(guchar * str, union sockunion *su);
boolean adl_equal_address(union sockunion *one, union sockunion *two);
const int adl_sockunion2str(union sockunion *su, guchar * buf, size_t len);


/* Defines the callback function that is called when new data have been received 
   (Events POLLIN||POLLPRI ) 
   Params: 1. socketdescriptor of the socket where the datagramm has been received 
           2. pointer to the datagramm data 
           3. length of datagramm data
           4. source Address
           5. length of source address
*/
typedef void (*Socket_callback_function) (Socket_descriptor, unsigned char *,
                                          int, union sockunion *, int);



/*------------------------------------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------------------------------------*/



/* Defines the callback function that is called when an timer expires.
   Params: 1. ID of timer
           2. pointer to param1
           3. pointer to param2
              param1 and param2 are transparent data from the caller of start_timer,
              that are returned to it when the timer expires.
*/
typedef void (*Timer_expires_callback) (TimerID, void *, void *);



/**
 * This function binds a local socket for incoming requests
 * @return socket file descriptor for the newly opened and bound socket
 * @param address (local) port to bind to
 */
Socket_descriptor adl_open_sctp_socket();

Socket_descriptor adl_get_sctpv4_socket(unsigned int portnum);
#ifdef HAVE_IPV6
Socket_descriptor adl_get_sctpv6_socket(unsigned int portnum);
#endif
Socket_descriptor adl_get_udp_socket(unsigned int portnum);


/**
 * function to be called when we get a message from a peer sctp instance in the poll loop
 * @param  sfd the socket file descriptor where data can be read...
 * @param  buf pointer to a buffer, where we data is stored
 * @param  len number of bytes to be sent, including the ip header !
 * @param  address, where data goes from
 * @param	dest_len size of the address
 * @return returns number of bytes actually sent, or error
 */
int adl_send_message(int sfd, void *buf, int len, union sockunion *dest);


/**
 * function add a sfd to the list of sfds we want to wait for with the poll()
 * @param sfd	        the socket file descriptor of the socket to react upon
 * @param scf           function pointer holding the callback funtion for normal events.
 * @return              the current number of sockets that are polled.
 */
int adl_register_socket_cb(Socket_descriptor sfd, Socket_callback_function scf);



/**
 * 	function to check for events on all poll fds (i.e. open sockets), or else
 * 	execute the next timer event. Executed timer events are removed from the list.
 *    Wrapper for poll()
 *	@return  0 after shutdown of the application, -1 for error
 *    @author  ajung
 */
int adl_event_loop();



/**
 * this function initializes the data of this module.
 */
void adl_init_adaptation_layer();

/**
 *      This function adds a callback that is to be called some time from now. It realizes
 *      the timer (in an ordered list).
 *      @param      milliseconds  action is to be started in milliseconds ms from now
 *      @param      timer_cb      pointer to a function to be executed, when timer expires
 *      @param      param1        pointer to be returned to the caller when timer expires
 *      @param      param2        pointer to be returned to the caller when timer expires
 *      @return     returns an id value, that can be used to cancel a timer
 *      @author     ajung
 */
TimerID adl_start_timer(unsigned int milliseconds,
                        Timer_expires_callback timer_cb, void *param1, void *param2);



/**
 *      This function stops a previously started timer.
 *      @param      tid        timer-id of timer to be removed
 *      @return     returns 0 on success, 1 if tid not in the list, -1 on error
 *      @author     ajung
 */
int adl_stop_timer(unsigned int tid);



/**
 *      Restarts a timer currently running
 *      @param      timer_id   the value returned by set_timer for a certain timer
 *      @param      milliseconds  action is to be started in milliseconds ms from now
 *      @return     new timer id , zero when there is an error (i.e. no timer)
 *      @author     ajung
 */

unsigned int adl_restart_timer_action(unsigned int timer_id, unsigned int milliseconds);



/**
 *	function to close a bound socket from our list of socket descriptors
 *	@param	sfd	socket file descriptor to be closed
 *	@return  0 on success, -1 for error, 1 if socket was not bound
 *    @author  ajung
 */
int adl_remove_cb(Socket_descriptor sfd);


/**
 * function is to return difference in msecs between time a and b (i.e. a-b)
 * @param a later time (e.g. current time)
 * @param b earlier time
 * @return -1 if a is earlier than b, else msecs that passed from b to a
 */
int adl_timediff_to_msecs(struct timeval *a, struct timeval *b);

/**
 * helper function for the sake of a cleaner interface :-)
 */
int adl_gettime(struct timeval *tv);

void adl_add_msecs_totime(struct timeval *t, unsigned int msecs);

#endif
