/*
 *  $Id: bundling.c,v 1.53 2000/11/02 16:17:04 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: This module bundels chunks chunks to be send into UDP-datagramms. Chunks are accepted
 *          with the putChunk function until sendChunk is called, which causes the transmission of
 *          all chunks accumulated so far.
 *          In the receive direction, this module gets UDP-dataparts from SCTP message-distribution,
 *          which are disassambled (debundled) into chunks. Depending on the chunk-type, the chunk
 *          are distributed to SCTP-control, RX-control, pathmanagement or reliable transfer.
 */

#include <stdio.h>
#include <stdlib.h>
#include "bundling.h"
#include "messages.h"
#include "dll_main.h"
#include "distribution.h"
#include "recvctrl.h"
#include "reltransfer.h"
#include "errorhandler.h"

#define TOTAL_SIZE(buf)		((buf)->ctrl_position+(buf)->sack_position+(buf)->data_position- 2*sizeof(SCTP_common_header))

typedef struct bundling_instance_struct
{

    guchar ctrl_buf[MAX_MTU_SIZE];
    guchar sack_buf[MAX_MTU_SIZE];
    guchar data_buf[MAX_MTU_SIZE];
    /* Leave some space for the SCTP common header */
    guint ctrl_position;
    guint sack_position;
    guint data_position;
    gboolean data_in_buffer;
    gboolean ctrl_chunk_in_buffer;
    gboolean sack_in_buffer;
    /* status flag for correct sequence of actions */
    gboolean treating_received_chunks;
    gboolean got_sack_request;
    gboolean got_send_request;
    gboolean got_send_address;
    gboolean got_sack;
    gboolean got_shutdown;
    guint requested_destination;
}
bundling_instance;

static bundling_instance *global_buffer;

void bu_init_bundling(void)
{
    global_buffer = bu_new();
}

gpointer bu_new(void)
{
    /* Alloc new bundling_instance data struct */
    bundling_instance *ptr;

    ptr = malloc(sizeof(bundling_instance));
    if (!ptr) {
        error_log(ERROR_MAJOR, "Malloc failed");
        return 0;
    }
    ptr->ctrl_position = sizeof(SCTP_common_header); /* start adding data after that header ! */
    ptr->data_position = sizeof(SCTP_common_header); /* start adding data after that header ! */
    ptr->sack_position = sizeof(SCTP_common_header); /* start adding data after that header ! */

    ptr->data_in_buffer = FALSE;
    ptr->ctrl_chunk_in_buffer = FALSE;
    ptr->sack_in_buffer = FALSE;
    return ptr;
}


void bu_delete(gpointer buPtr)
{
    event_log(INTERNAL_EVENT_0, "deleting bundling");
    free(buPtr);
}



gboolean bu_scanDatagram(guchar * datagram, guint len, gushort chunk_type)
{
    gushort processed_len = 0;
    guchar *current_position;
    guint pad_bytes;
    SCTP_simple_chunk *chunk;

    current_position = datagram; /* points to the first chunk in this pdu */
    while (processed_len < len) {

        event_logii(INTERNAL_EVENT_0,
                    "bu_scanDatagram : len==%u, processed_len == %u", len, processed_len);

        chunk = (SCTP_simple_chunk *) current_position;
        switch (chunk->chunk_header.chunk_id == chunk_type) {
        case TRUE:
            return TRUE;
            break;
        default:
            processed_len += CHUNKP_LENGTH((SCTP_chunk_header *) chunk);
            pad_bytes = ((processed_len % 4) == 0) ? 0 : (4 - processed_len % 4);
            processed_len += pad_bytes;
            current_position +=
                (CHUNKP_LENGTH((SCTP_chunk_header *) chunk) + pad_bytes * sizeof(unsigned char));
            break;
        }
    }
    return FALSE;
}


guchar* bu_findChunk(guchar * datagram, guint len, gushort chunk_type)
{
    gushort processed_len = 0;
    guchar *current_position;
    guint pad_bytes;
    SCTP_simple_chunk *chunk;

    current_position = datagram; /* points to the first chunk in this pdu */
    while (processed_len < len) {

        event_logii(INTERNAL_EVENT_0,
                    "bu_findChunk : len==%u, processed_len == %u", len, processed_len);

        chunk = (SCTP_simple_chunk *) current_position;
        switch (chunk->chunk_header.chunk_id == chunk_type) {
        case TRUE:
            return current_position;
            break;
        default:
            processed_len += CHUNKP_LENGTH((SCTP_chunk_header *) chunk);
            pad_bytes = ((processed_len % 4) == 0) ? 0 : (4 - processed_len % 4);
            processed_len += pad_bytes;
            current_position +=
                (CHUNKP_LENGTH((SCTP_chunk_header *) chunk) + pad_bytes * sizeof(unsigned char));
            break;
        }
    }
    return NULL;
}

gboolean bu_scanDatagramForError(guchar * datagram, guint len, gushort error_cause)
{
    gushort processed_len = 0;
    gushort err_len = 0;

    guchar *current_position;
    guint pad_bytes;
    SCTP_simple_chunk *chunk;
    SCTP_staleCookieError *err_chunk;


    current_position = datagram; /* points to the first chunk in this pdu */
    while (processed_len < len) {

        event_logii(VERBOSE,
                    "bu_scanDatagramForError : len==%u, processed_len == %u", len, processed_len);

        chunk = (SCTP_simple_chunk *) current_position;
        switch (chunk->chunk_header.chunk_id == CHUNK_ERROR) {
        case TRUE:
            event_log(INTERNAL_EVENT_0, "bu_scanDatagramForError : Error Chunk Found");
            while (err_len < chunk->chunk_header.chunk_length) {
                err_chunk = (SCTP_staleCookieError *) & current_position[err_len];
                if (ntohs(err_chunk->vlparam_header.param_type) == error_cause) {
                    event_logi(VERBOSE,
                               "bu_scanDatagramForError : Error Cause %u found -> Returning TRUE",
                               error_cause);
                    return TRUE;
                }
                err_len += ntohs(err_chunk->vlparam_header.param_length);
                while ((err_len % 4) != 0)
                    err_len++;
            }
        default:
            processed_len += CHUNKP_LENGTH((SCTP_chunk_header *) chunk);
            pad_bytes = ((processed_len % 4) == 0) ? 0 : (4 - processed_len % 4);
            processed_len += pad_bytes;
            current_position +=
                (CHUNKP_LENGTH((SCTP_chunk_header *) chunk) + pad_bytes * sizeof(unsigned char));
            break;
        }
    }
    event_logi(VERBOSE,
               "bu_scanDatagramForError : Error Cause %u NOT found -> Returning FALSE",
               error_cause);
    return FALSE;


}

gint bu_rcvDatagram(guint address_index, guchar * datagram, guint len)
{
    /* sctp common header header has been verified */
    /* tag (if association is established) and CRC is okay */
    /* get first chunk-id and length, pass pointers & len on to relevant module :
       - CHUNK_INIT, CHUNK_INIT_ACK,CHUNK_ABORT, CHUNK_SHUTDOWN,CHUNK_SHUTDOWN_ACK
       CHUNK_COOKIE_ECHO,CHUNK_COOKIE_ACK go to SCTP_CONTROL (change of association state)
       - CHUNK_HBREQ, CHUNK_HBACK go to PATH_MAN instance
       - CHUNK_SACK goes to RELIABLE_TRANSFER
       - CHUNK_ERROR probably to SCTP_CONTROL as well  (at least there !)
       - CHUNK_DATA goes to RX_CONTROL
     */
    guchar *current_position;
    gushort processed_len = 0, chunk_len;
    gushort pad_bytes;
    SCTP_simple_chunk *chunk;
    gboolean data_chunk_received = FALSE, send_it;
    bundling_instance *bu_ptr;

    bu_ptr = (bundling_instance *) mdi_readBundling();
    if (!bu_ptr) {              /* Assume that no association exists, so we take the global bundling buffer */
        event_log(INTERNAL_EVENT_0, "Setting global bundling buffer ");
        bu_ptr = global_buffer;
    }
    bu_ptr->treating_received_chunks = TRUE;
    bu_ptr->got_sack_request = FALSE;
    bu_ptr->got_sack = FALSE;
    bu_ptr->got_send_request = FALSE;
    bu_ptr->got_send_address = FALSE;
    bu_ptr->got_shutdown = FALSE;

    current_position = datagram; /* points to the first chunk in this pdu */
    event_log(INTERNAL_EVENT_0, "Entered Bundling...... ");
    /* CHECKME : beim Empfangen leerer Chunks tritt im Bundling eine Endlosschleife auf ??? */
    while (processed_len < len) {

        chunk = (SCTP_simple_chunk *) current_position;
        chunk_len = CHUNKP_LENGTH((SCTP_chunk_header *) chunk);
        event_logiiii(INTERNAL_EVENT_0,
                     "bu_rcvDatagram(address=%u) : len==%u, processed_len = %u, chunk_len=%u",
                     address_index, len, processed_len, chunk_len);
        if (chunk_len > len) {
            error_logii(ERROR_MINOR, "Faulty chunklen=%u, total len=%u --> dropping rest of data !",
                                    chunk_len,len);

            bu_ptr->treating_received_chunks = FALSE;
            return 1;
        }
        /*
         * TODO :   Add return values to the chunk-functions, where they can indicate what
         *          to do with the rest of the datagram (i.e. DISCARD after stale COOKIE_ECHO
         *          with tie tags that do not match the current ones)
         */
        switch (chunk->chunk_header.chunk_id) {
        case CHUNK_DATA:
            event_log(INTERNAL_EVENT_0, "Bundling received DATA chunk");
            rxc_data_chunk_rx((SCTP_data_chunk *) chunk, address_index);
            data_chunk_received = TRUE;
            break;
        case CHUNK_INIT:
            event_log(INTERNAL_EVENT_0, "Bundling received INIT chunk");
            scr_init((SCTP_init *) chunk);
            break;
        case CHUNK_INIT_ACK:
            event_log(INTERNAL_EVENT_0, "Bundling received INIT ACK chunk");
            scr_initAck((SCTP_init *) chunk);
            break;
        case CHUNK_SACK:
            event_log(INTERNAL_EVENT_0, "Bundling received SACK chunk");
            bu_ptr->got_sack = TRUE;
            rtx_process_sack(address_index, chunk);
            bu_ptr->got_sack = FALSE;
            break;
        case CHUNK_HBREQ:
            event_log(INTERNAL_EVENT_0, "Bundling received HB_REQ chunk");
            pm_heartbeat((SCTP_heartbeat *) chunk, address_index);
            break;
        case CHUNK_HBACK:
            event_log(INTERNAL_EVENT_0, "Bundling received HB_ACK chunk");
            pm_heartbeatAck((SCTP_heartbeat *) chunk);
            break;
        case CHUNK_ABORT:
            event_log(INTERNAL_EVENT_0, "Bundling received ABORT chunk");
            bu_ptr->got_shutdown = TRUE;
            scr_abort();
            break;
        case CHUNK_SHUTDOWN:
            event_log(INTERNAL_EVENT_0, "Bundling received SHUTDOWN chunk");
            bu_ptr->got_shutdown = TRUE;
            scr_shutdown((SCTP_simple_chunk *) chunk);
            break;
        case CHUNK_SHUTDOWN_ACK:
            event_log(INTERNAL_EVENT_0, "Bundling received SHUTDOWN ACK chunk");
            bu_ptr->got_shutdown = TRUE;
            scr_shutdownAck();
            break;
        case CHUNK_ERROR:
            event_log(INTERNAL_EVENT_0, "Error Chunk");
            eh_recv_chunk(chunk);
            break;
        case CHUNK_COOKIE_ECHO:
            event_log(INTERNAL_EVENT_0, "Bundling received COOKIE ECHO chunk");
            scr_cookie_echo((SCTP_cookie_echo *) chunk);
            break;
        case CHUNK_COOKIE_ACK:
            event_log(INTERNAL_EVENT_0, "Bundling received COOKIE ACK chunk");
            scr_cookieAck((SCTP_simple_chunk *) chunk);
            break;
        case CHUNK_ECNE:
        case CHUNK_CWR:
            event_logi(INTERNAL_EVENT_0,
                       "Chunktype %u not Supported Yet !!!!!!!!", chunk->chunk_header.chunk_id);
            break;
        case CHUNK_SHUTDOWN_COMPLETE:
            event_log(INTERNAL_EVENT_0, "Bundling received SHUTDOWN_COMPLETE chunk");
            scr_shutdownComplete();
            break;
        default:
            event_log(EXTERNAL_EVENT_X, "Received unknown chunk type in bundling.c");
            break;
        }

        processed_len += chunk_len;
        pad_bytes = ((processed_len % 4) == 0) ? 0 : (4 - processed_len % 4);
        processed_len += pad_bytes;
        current_position += (chunk_len + pad_bytes) * sizeof(unsigned char);

        event_logiiii(VVERBOSE, "processed_len=%u, pad_bytes=%u, current_position=%u, chunk_len=%u",
            processed_len, pad_bytes, current_position,chunk_len);
    }
    if (data_chunk_received == TRUE) {
        /* update SACK structure */
        rxc_all_chunks_processed();
    }

    bu_ptr->treating_received_chunks = FALSE;


    /* if (bu_ptr->got_shutdown == TRUE)  then chunks have already been sent... */
    if (bu_ptr->got_shutdown == FALSE) {
        if (bu_ptr->got_sack_request == TRUE || data_chunk_received == TRUE) {
            send_it = rxc_create_sack(&address_index, FALSE);
            if (send_it == TRUE)
                bu_sendAllChunks(&address_index);
        } else if (bu_ptr->got_send_request == TRUE || bu_ptr->got_sack_request == TRUE) {
            if (bu_ptr->got_send_address == TRUE) {
                bu_sendAllChunks(&(bu_ptr->requested_destination));
            } else {
                bu_sendAllChunks(NULL);
            }
        }
    }
    /* resetting flags */
    bu_ptr->got_sack_request = FALSE;
    bu_ptr->got_send_request = FALSE;
    bu_ptr->got_send_address = FALSE;

    return 0;

}

void bu_request_sack()
{
    bundling_instance *bu_ptr;
    event_log(INTERNAL_EVENT_0, "bu_request_sack() was called... ");

    bu_ptr = (bundling_instance *) mdi_readBundling();
    if (!bu_ptr) {              /* Assume that no association exists, so we take the global bundling buffer */
        event_log(INTERNAL_EVENT_0, "Setting global bundling buffer ");
        bu_ptr = global_buffer;
    }
    bu_ptr->got_sack_request = TRUE;

}

/* TODO : Handle multiple calls to this function between two send events */
gint bu_put_SACK_Chunk(SCTP_sack_chunk * chunk)
{
    gint result;
    bundling_instance *bu_ptr;

    event_log(INTERNAL_EVENT_0, "bu_put_SACK_Chunk() was called ");

    bu_ptr = (bundling_instance *) mdi_readBundling();

    if (!bu_ptr) {              /* Assume that no association exists, so we take the global bundling buffer */
        event_log(INTERNAL_EVENT_0, "Copying data to global bundling buffer ");
        bu_ptr = global_buffer;
    }
    if (TOTAL_SIZE(bu_ptr) + CHUNKP_LENGTH((SCTP_chunk_header *) chunk) >= MAX_SCTP_PDU) {
        error_log(ERROR_MAJOR,
                  "Chunk Length exceeded MAX_SCTP_PDU : sending chunk to default address !");
        result = bu_sendAllChunks(NULL);
    }
    if (bu_ptr->sack_in_buffer == TRUE) { /* multiple calls in between */
        event_log(INTERNAL_EVENT_0,
                  "bu_put_SACK_Chunk was called a second time, deleting first chunk");
        bu_ptr->sack_position = sizeof(SCTP_common_header);
    }

    memcpy(&(bu_ptr->sack_buf[bu_ptr->sack_position]), chunk,
           CHUNKP_LENGTH((SCTP_chunk_header *) chunk));
    /* SACK always multiple of 32 bytes, do not care about padding */
    bu_ptr->sack_position += CHUNKP_LENGTH((SCTP_chunk_header *) chunk);
    bu_ptr->sack_in_buffer = TRUE;
    return 0;
}

/* this function used for sending control chunks, so it was renamed */
gint bu_put_Ctrl_Chunk(SCTP_simple_chunk * chunk)
{
    gint result;
    bundling_instance *bu_ptr;
    gint count;

    event_log(INTERNAL_EVENT_0, "bu_put_Ctrl_Chunk() was called");

    bu_ptr = (bundling_instance *) mdi_readBundling();

    if (!bu_ptr) {              /* Assume that no association exists, so we take the global bundling buffer */
        event_log(INTERNAL_EVENT_0, "Copying data to global bundling buffer ");
        bu_ptr = global_buffer;
    }

    if (TOTAL_SIZE(bu_ptr) + CHUNKP_LENGTH((SCTP_chunk_header *) chunk) >= MAX_SCTP_PDU) {
        error_log(ERROR_MAJOR,
                  "Chunk Length exceeded MAX_SCTP_PDU : sending chunk to default address !");
        result = bu_sendAllChunks(NULL);
    }
    memcpy(&(bu_ptr->ctrl_buf[bu_ptr->ctrl_position]), chunk,
           CHUNKP_LENGTH((SCTP_chunk_header *) chunk));

    bu_ptr->ctrl_position += CHUNKP_LENGTH((SCTP_chunk_header *) chunk);
    /* insert padding, if necessary */
    if ((CHUNKP_LENGTH((SCTP_chunk_header *) chunk) % 4) != 0) {
        for (count = 0; count < (4 - (CHUNKP_LENGTH((SCTP_chunk_header *) chunk) % 4)); count++) {
            bu_ptr->ctrl_buf[bu_ptr->ctrl_position] = 0;
            bu_ptr->ctrl_position++;
        }
    }
    bu_ptr->ctrl_chunk_in_buffer = TRUE;
    return 0;
}

gint bu_put_Data_Chunk(SCTP_simple_chunk * chunk)
{
    gint result;
    bundling_instance *bu_ptr;
    gint count;

    event_log(INTERNAL_EVENT_0, "bu_put_Data_Chunk() was called ");

    bu_ptr = (bundling_instance *) mdi_readBundling();

    if (!bu_ptr) {              /* Assume that no association exists, so we take the global bundling buffer */
        event_log(INTERNAL_EVENT_0, "Copying data to global bundling buffer ");
        bu_ptr = global_buffer;
    }

    if (TOTAL_SIZE(bu_ptr) + CHUNKP_LENGTH((SCTP_chunk_header *) chunk) >= MAX_SCTP_PDU) {
        error_log(ERROR_MAJOR,
                  "Chunk Length exceeded MAX_SCTP_PDU : sending chunk to default address !");
        result = bu_sendAllChunks(NULL);
    }
    memcpy(&(bu_ptr->data_buf[bu_ptr->data_position]), chunk,
           CHUNKP_LENGTH((SCTP_chunk_header *) chunk));
    bu_ptr->data_position += CHUNKP_LENGTH((SCTP_chunk_header *) chunk);

    event_logii(VERBOSE,
                "Put Data Chunk Length : %u , Total buffer size : %u\n",
                CHUNKP_LENGTH((SCTP_chunk_header *) chunk), TOTAL_SIZE(bu_ptr));

    /* insert padding, if necessary */
    if ((CHUNKP_LENGTH((SCTP_chunk_header *) chunk) % 4) != 0) {
        for (count = 0; count < (4 - (CHUNKP_LENGTH((SCTP_chunk_header *) chunk) % 4)); count++) {
            bu_ptr->data_buf[bu_ptr->data_position] = 0;
            bu_ptr->data_position++;
        }
    }
    event_log(INTERNAL_EVENT_0, "bu_put_Data_Chunk() : Added Padding bytes");
    bu_ptr->data_in_buffer = TRUE;
    return 0;
}


gint bu_sendAllChunks(guint * ad_idx)
{
    gint result, send_len = 0;
    guchar *send_buffer = NULL;
    bundling_instance *bu_ptr;
    gshort idx = 0;

    bu_ptr = (bundling_instance *) mdi_readBundling();

    event_log(INTERNAL_EVENT_0, "bu_sendAllChunks() is being executed...");

    if (!bu_ptr) {
        event_log(INTERNAL_EVENT_0, "Sending data from global bundling buffer ");
        bu_ptr = global_buffer;
    }
    event_log(VERBOSE, "bu_sendAllChunks : Marke 0");
    if (bu_ptr->treating_received_chunks == TRUE && bu_ptr->got_sack == FALSE) {
        bu_ptr->got_send_request = TRUE;
        if (ad_idx) {
            bu_ptr->got_send_address = TRUE;
            bu_ptr->requested_destination = *ad_idx;
        }
        if (bu_ptr->got_shutdown == FALSE) {
            event_log(VERBOSE, "bu_sendAllChunks : returning 0");
            return 0;
        }
    }

    /* TODO : more intelligent path selection strategy */
    /*         should take into account PM_INACTIVE */
    if (ad_idx) {
        if (*ad_idx > 0xFFFF)
            error_log(ERROR_FATAL, "address_index too big !");
        else
            idx = (short) *ad_idx;
        if (pm_readState(idx) == PM_INACTIVE) {
            event_logi(INTERNAL_EVENT_0, "Trying to send chunks over inactive interface %d", idx);
        }
    } else {
        idx = pm_readPrimaryPath();
        if (pm_readPrimaryPath() == -1) {
            error_log(ERROR_MINOR,
                      "Primary Path returned was -1 in bu_sendAllChunks(), using PATH 0");
            idx = 0;
        }
        /* the last resort...... */
        if (pm_readState(idx) == PM_INACTIVE)
            idx = rxc_read_last_active_address();
    }

    if (bu_ptr->sack_in_buffer)
        send_buffer = bu_ptr->sack_buf;
    else if (bu_ptr->ctrl_chunk_in_buffer)
        send_buffer = bu_ptr->ctrl_buf;
    else if (bu_ptr->data_in_buffer)
        send_buffer = bu_ptr->data_buf;
    else
        error_log(ERROR_MINOR, "Nothing to send, but bu_sendAllChunks was called !");

    if (bu_ptr->sack_in_buffer) {
        rxc_stop_sack_timer();
        /* SACKs by default go to the last active address, from which data arrived */
        idx = rxc_read_last_active_address();
        send_len = bu_ptr->sack_position; /* at least sizeof(SCTP_common_header) */
        /* at most pointing to the end of SACK chunk */
        if (bu_ptr->ctrl_chunk_in_buffer) {
            memcpy(&send_buffer[send_len],
                   &(bu_ptr->ctrl_buf[sizeof(SCTP_common_header)]),
                   (bu_ptr->ctrl_position - sizeof(SCTP_common_header)));
            send_len += bu_ptr->ctrl_position - sizeof(SCTP_common_header);
        }
        if (bu_ptr->data_in_buffer) {
            memcpy(&send_buffer[send_len],
                   &(bu_ptr->data_buf[sizeof(SCTP_common_header)]),
                   (bu_ptr->data_position - sizeof(SCTP_common_header)));
            send_len += bu_ptr->data_position - sizeof(SCTP_common_header);
        }
    } else if (bu_ptr->ctrl_chunk_in_buffer) {
        send_len = bu_ptr->ctrl_position;
        if (bu_ptr->data_in_buffer) {
            memcpy(&send_buffer[send_len],
                   &(bu_ptr->data_buf[sizeof(SCTP_common_header)]),
                   (bu_ptr->data_position - sizeof(SCTP_common_header)));
            send_len += bu_ptr->data_position - sizeof(SCTP_common_header);
        }

    } else if (bu_ptr->data_in_buffer)
        send_len = bu_ptr->data_position;

    if (bu_ptr->data_in_buffer) pm_chunksSentOn(idx);

    event_logii(VERBOSE,
                "bu_sendAllChunks() : sending message len==%u to adress idx=%u", send_len, idx);
    result = mdi_send_message((SCTP_message *) send_buffer, send_len, idx);
    event_logi(VVERBOSE, "bu_sendAllChunks(): result==%d ", result);
    /* reset all positions */
    bu_ptr->sack_in_buffer = FALSE;
    bu_ptr->ctrl_chunk_in_buffer = FALSE;
    bu_ptr->data_in_buffer = FALSE;

    bu_ptr->data_position = sizeof(SCTP_common_header);
    bu_ptr->ctrl_position = sizeof(SCTP_common_header);
    bu_ptr->sack_position = sizeof(SCTP_common_header);
    return result;
}
