/*
 *  $Id: distribution.c,v 1.78 2000/11/02 16:48:24 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: This modules implements the interface defined distribution.h and sctp.h
 *          and holds a private list of associations.
 *          As an SCTP-instance has usually more than one associations, the main purpose
 *          of this module is the  distribution of signals from the ULP and from the 
 *          peer (via the socket and unix-interface) to the addressed association.
 *          Signals from the UNIX-interface are always forwarded to 
 *          the bundling module of the addressed association.
 *          The corresponding function receive_msg is not defined here but in distribution.c,
 *          because it is called via a function-pointer that is registered at the
 *          UNIX-interface when SCTP is initialized.
 *          For incoming UDP-datagramms, the tag-value in datagram is used to identify the
 *          association.
 *          Signals from the ULP are forwarded to SCTP-Control, Pathmanagement or
 *          Streamengine of the addressed association depending on the chunk type.
 *          All signals from the ULP contain the association ID, which is used to
 *          identify the association.
 *
 * Remarks: Host and network byte order (HBO and NBO):
 *          In this module all data are data are HBO, except the IP-Addresses and message strings.
 *          Thus everything entered into a message string must be converted to NBO and
 *          everthing extracted from a message string must be converted to HBO (except the
 *          IP-addresses).
 *
 * function prefixes: sctp_    for interface-defs imported from sctp.h
 *                    mdi_     for interface-defs imported from distribution.h
 */

#include  "dll_main.h"          /* doubly linked list library */
#include  "sctp.h"              /* ULP-interface definitions */
#include  "distribution.h"      /* SCTP-internal interfaces to message distribution */
#include  "adaptation.h"        /* interfaces to adaptation layer */
#include  "bundling.h"          /* interfaces to bundling */
#include  "SCTP-control.h"      /* interfaces to SCTP-control */
#include  "auxiliary.h"
#include  "streamengine.h"      /* interfaces to streamengine */
#include  "flowcontrol.h"       /* interfaces to flowcontrol */
#include  "recvctrl.h"          /* interfaces to receive-controller */
#include  "chunkHandler.h"

#include  <sys/types.h>
#include  <stdlib.h>            /* for malloc */
#include  <stdio.h>             /* for malloc */
#include  <errno.h>
#include  <arpa/inet.h>         /* for inet_ntoa() under both SOLARIS/LINUX */

/**************************************************************************************************/
/* function to handle typed user commands */
/* FOR THE BAKEOFF */
void
ulp_stdin_cb(Socket_descriptor sin_fd, unsigned char *buffer,
             int bufferLength, union sockunion *fromAddress, int fromAddressLength);
void
ulp_socket_error(Socket_descriptor sin_fd,
                 unsigned char *buffer,
                 int bufferLength, union sockunion *fromAddress, int fromAddressLength);

/**************************************************************************************************/
extern gushort sctp_instance;
extern gboolean support_ipv6;

/******************** Structure Definitions *******************************************************/

/* This struct stores data of SCTP-instances.
   Each SCTP-instances is related to one port and to
   one SCTP adaption-layer.
*/
typedef struct SCTPINSTANCE
{
    unsigned short sctpInstanceName; /* The name of this SCTP-instance, used as key. */
    unsigned short localPort;   /* The local port of this SCTP-instance where inits
                                   are accepted. */
    guint16 noOfLocalAddresses;
    union sockunion *localAddressList;
    struct ULPcallback ULPcallbackFunctions;
    boolean ulp_via_udp;        /* is true if we communicate via UDP to ULP */
    gushort udp_peer_port;      /* udp port used for talking to ULP */
}
SCTP_instance;


/* This struct contains all data of an association. As far as other modules must know elements
   of this struct, read functions are provided. No other module has write access to this structure.
*/
typedef struct ASSOCIATION
{
    unsigned int tagLocal;      /* The local tag of this assocaition.
                                   It is also used as ID of this association, 
                                   it is used as a key to find a association in
                                   the list.  */
    unsigned int tagRemote;     /* The tag of remote side of this association */

    unsigned short sctpInstanceName; /* The name of the SCTP-instance this association
                                        belongs to. It is equal to the wellknown port
                                        number of the ULP that uses this instance. */

    unsigned short localPort;   /* the local port number of this association. */

    unsigned short remotePort;  /* the remote port number of this association. */

    unsigned int noOfNetworks;  /* number of destination networks (paths) */

    union sockunion *destinationAddresses; /* array of destination addresses */

    /* pointer to other modules of SCTP */
    void *flowControl;          /* pointer to flowcontrol structure */
    void *reliableTransfer;     /* pointer to reliable-transfer structure */
    void *rx_control;           /* pointer to receive-control structure */
    void *streamengine;         /* pointer to stream structure */
    void *pathMan;              /* pointer to pathmanagement structure */
    void *bundling;             /* pointer to bundling structure */
    void *sctp_control;         /* pointer to SCTP-control */

    boolean deleted;            /* marks an association for deletion */
} Association;


/******************** Declarations ****************************************************************/

/*
    Keyed list of SCTP-instances with the instanceName as key
*/
/* TODO : ??? static List *sctpInstances; ??? if more than one listening port. */

/* Keyed list of associations with the association-ID as key
*/
static List *associations;


/* Temporary Temporary Temporary Temporary Temporary Temporary Temporary Temporary Temporary */
/* as long there is only one sctp-instance */

static SCTP_instance *sctpInstance;

/* Temporary Temporary Temporary Temporary Temporary Temporary Temporary Temporary Temporary */





/* Whenever an external event (ULP-call, socket-event or timer-event) this variable must
   contain the addressed association.
   Read functions for 'global data' read data from the association pointed to by this pointer.
   This pointer must be reset to null after the event  has been handled.
*/
static Association *currentAssociation;
static Association tmpAssoc;
static union sockunion tmpAddress;

/* Asssociation ID's are derived by a random number generator because they are also
   used as local tag. If firstSCTP_instance is true, a seed is generated by proper
   means (current time). After the first SCTP-instance was created, firstSCTP_instance
   is set to false.
*/
static boolean clearAssociation;
static boolean firstSCTP_instance = TRUE;
static unsigned short lastSCTP_instanceName = 1;

/*
 *  initAck is sent to this address
 *  In this case, SCTP-control reads this address on reception of the cookie echo (which consequently
 *  also does not contain an addresslist) to initialize the new association.
 */
static union sockunion *lastFromAddress;

static short lastFromPath;
static unsigned short lastFromPort;
static unsigned short lastDestPort;
static unsigned int lastInitiateTag;

/* Descriptor of socket used by all associations and SCTP-instances.
*/
static Socket_descriptor socket_descriptor;
#ifdef HAVE_IPV6
static Socket_descriptor ipv6_socket_descriptor;
#endif
static Socket_descriptor udp_socket_descriptor;


/* port management array */
static unsigned char *portsSeized;
static unsigned int lastPortSeized;
static unsigned int numberOfSeizedPorts;

static long rstate[2];


/******************** Internal Function Prototypes ************************************************/

static void mdi_deleteAssociation(Association * assoc);

/******************** Function Implementations ****************************************************/

/*------------------- Internal Functions ---------------------------------------------------------*/




/*------------------- Internal LIST Functions ----------------------------------------------------*/

/* equalAssociationIDs compares the association ID's of two associations and returns 0
   if they are equal. This is a call back function called by dll_list whenever two
   association need to be compared.
   @param il1 : association data 1
   @param il2 : association data 2
   return:      0 if il1->tagLocal equals il2->tagLocal, 1 if bigger, -1 if smaller
*/
int equalAssociationIDs(Info * i1, Info * i2)
{
    Association *a1, *a2;

    a1 = *((Association **) i1);
    a2 = *((Association **) i2);

    /* delete association that have already been marked for deletion */
    /* a1 is the one in list, a2 is the one to be matched */
    if (a1->deleted) {
        return 0;
    }

    /* two associations are equal if there local tags (in this implementation also used as
       association ID) are equal. */
    if (a1->tagLocal == a2->tagLocal)
        return 0;
    else if (a1->tagLocal < a2->tagLocal)
        return -1;
    else
        return 1;
}



/* equalAssociations compares two associations and returns 0 if they are equal. In contrast to
   function equalAssociationIDs, equal here means the two associations belong to the same
   SCTP-instance and have at least one destinationaddress in common.
   This is a call back function called by dll_list whenever two association need to be compared.
   @param il1 : association data 1
   @param il2 : association data 2
   return:      0 if il1 and il2 are equal according to above definition, 1 else
*/
int equalAssociations(Info * i1, Info * i2)
{
    Association *a1, *a2;
    int i, j;

    a1 = *((Association **) i1);
    a2 = *((Association **) i2);

    /* two associations are equal if there remote ports are equal and at least 
       one of there remote addresses are equal. This is like in TCP, where a connection
       is identified by the transport address, i.e. the IP-address and port of the peer. */
    if (a1->remotePort == a2->remotePort) {
        for (i = 0; i < a1->noOfNetworks; i++)
            for (j = 0; j < a2->noOfNetworks; j++) {
                if (adl_equal_address
                    (&(a1->destinationAddresses[i]), &(a2->destinationAddresses[j])) == TRUE)
                    return 0;
            }
        return 1;
    }
    return 1;
}



/* enterAssociation enters a association into the list.
   @param assoc : pointer to association data
*/
static void enterAssociation(Association * assoc)
{
    DLL_Return exitCode;

    event_logi(INTERNAL_EVENT_0, "entering association %08x into list", assoc->tagLocal);

    if ((exitCode = DLL_AddRecord(associations, (Info *) & assoc, NULL)) != DLL_NORMAL)
        DLL_error_log(ERROR_FATAL, exitCode);
}



/* retrieveAssociation retrieves a association from the list using the local tag as key..
   @param assocID : association ID
   @return Association : the retrieved association
*/
Association *retrieveAssociation(unsigned int assocID)
{
    Association *assoc;
    Association *assocFindP;
    DLL_Return exitCode;

    event_logi(INTERNAL_EVENT_0, "retrieving association %08x from list", assocID);

    tmpAssoc.tagLocal = assocID;
    tmpAssoc.deleted = FALSE;
    assocFindP = &tmpAssoc;
    assoc = NULL;

    exitCode =
        DLL_FindRecord(associations, (Info *) & assoc, (Info *) & assocFindP, &equalAssociationIDs);

    if (exitCode == DLL_NORMAL) {
        if (assoc->deleted) {
            /* remove the association from the list */
            DLL_DeleteCurrentRecord(associations);
            /* free all association data */
            mdi_deleteAssociation(assoc);
            assoc = NULL;
            /* continue search for the addressed association */
        }
    } else if (exitCode == DLL_NOT_FOUND || exitCode == DLL_NULL_LIST) {
        event_logi(INTERNAL_EVENT_0, "association %08x not in list", assocID);
        assoc = NULL;
    } else {
        DLL_error_log(ERROR_FATAL, exitCode);
        assoc = NULL;
    }
    return assoc;
}



/* retrieveAssociationULP retrieves a association from the list using the local tag as key.
   This function is required, because for ULP calls at SCTP, it is not known if the association
   pointer is set or not. For instance sctp_sendChunk can result from an event originating from SCTP
   or for instance from a time event. In the first case the association must not be retrieved
   from the list, whereas in second it must.   
   retrieveAssociationULP bypasses this problems by retrieving the association only if
   currentAssociation equals NULL. Furthermore the global parameter clearAssociation is
   set accordingly to signal the caller it should clear the currentAssociation after the
   ULP call has been handled.
   @param assocID : association ID
   @return Association : the retrieved association
*/
Association *retrieveAssociationULP(unsigned int assocID)
{
    Association *assoc;

    /* Retrieve association from list  */
    if (currentAssociation == NULL) {
        assoc = retrieveAssociation(assocID);
        clearAssociation = TRUE;
    } else {
        if (currentAssociation->tagLocal != assocID) {
            /* Darf nicht vorkommen */
            error_log(ERROR_FATAL, "retrieveAssociationULP: Assoc. set but ID's do not match");
        }
        clearAssociation = FALSE;
        assoc = currentAssociation;
    }
    return assoc;
}



/*
    CHECKME : Must return NULL, if no Address-Port combination does not occur in ANY existing assoc.
              If it occurs in one of these -> return it

    retrieveAssociation retrieves a association from the list using the transport address as key.
    @param assocID : association ID
    @return Association : the retrieved association
*/
Association *retrieveAssociationByTransportAddress(union sockunion * fromAddress,
                                                   unsigned short fromPort)
{
    DLL_Return exitCode;
    Association *assocr;
    Association *assocp;

    tmpAssoc.noOfNetworks = 1;
    tmpAssoc.destinationAddresses = &tmpAddress;
    switch (sockunion_family(fromAddress)) {
    case AF_INET:
        event_logi(INTERNAL_EVENT_0,
                   "Looking for IPv4 Address %x, check NTOHX() ! ", sock2ip(fromAddress));
        tmpAssoc.destinationAddresses[0].sa.sa_family = AF_INET;
        tmpAssoc.destinationAddresses[0].sin.sin_addr.s_addr = sock2ip(fromAddress);
        tmpAssoc.remotePort = fromPort;
        tmpAssoc.deleted = FALSE;
        break;
#ifdef HAVE_IPV6
    case AF_INET6:
        tmpAssoc.destinationAddresses[0].sa.sa_family = AF_INET6;
        event_log(INTERNAL_EVENT_0, "Looking for IPv6 Address %x, check NTOHX() ! ");
        memcpy(tmpAssoc.destinationAddresses[0].sin6.sin6_addr.s6_addr,
               sock2ip6(fromAddress), sizeof(struct in6_addr));
        tmpAssoc.remotePort = fromPort;
        tmpAssoc.deleted = FALSE;
        break;
#endif
    default:
        error_logi(ERROR_FATAL,
                   "Unsupported Address Type %d in retrieveAssociationByTransportAddress()",
                   sockunion_family(fromAddress));
        break;

    }

    assocp = &tmpAssoc;

    event_log(INTERNAL_EVENT_0, "retrieving association by transport address from list");

    exitCode =
        DLL_FindRecord(associations, (Info *) & assocr, (Info *) & assocp, &equalAssociations);

    if (exitCode == DLL_NORMAL)
        return assocr;
    else if (exitCode == DLL_NOT_FOUND || exitCode == DLL_NULL_LIST) {
        event_log(INTERNAL_EVENT_0, "association indexed by transport address not in list");
        return NULL;
    } else {
        DLL_error_log(ERROR_FATAL, exitCode);
        return NULL;
    }
}



/* checkForExistingAssociations checks wether a given association is already in the list using
   the equality condition given by function equalAssociations.
   @param assoc_new : the association to be compared with the association in the list.
   @return found :    1 association found 0 else
*/
static short checkForExistingAssociations(Association * assoc_new)
{
    DLL_Return exitCode;

    /* get all Associations from the list that have the same ports ?! */

    /*  check if these have same number of transport addresses, and if these
       are equal for both associations. If so, the two associations are equal */
    event_log(INTERNAL_EVENT_1,
              "TODO : Function checkForExistingAssociations() not correctly implemented yet !");

    exitCode = 0;
    return exitCode;
}



/*------------------- Internal port management Functions -----------------------------------------*/

/* seizePort return a free port.
   @return port : free port.
*/
static unsigned short seizePort()
{
    /* this is not nice, as you can guess the next port that will be chosen quite easily */
    /* maybe  replace that by a sort of bit mapped array, marking free/used ports,  that */
    /* we can choose randomly from ?!   */
    if (numberOfSeizedPorts >= 0xFBFF)
        return 0x0000;

    while (portsSeized[++lastPortSeized])
        if (lastPortSeized >= 0xFFFF)
            lastPortSeized = 0x00000400;

    numberOfSeizedPorts++;
    portsSeized[lastPortSeized] = 1;

    if (lastPortSeized > 0xFFFF)
        error_log(ERROR_MINOR, "Warning: lastPortSeized exceeds range of unsigned short");

    return (unsigned short) lastPortSeized;
}



/* releasePort frees a used port.
   @param portSeized : freed port.
*/
static void releasePort(unsigned char portSeized)
{
    if (portsSeized[portSeized])
        error_log(ERROR_MINOR, "Warning: release of port that is not seized");

    numberOfSeizedPorts--;
    portsSeized[portSeized] = 0;
}



/*------------------- Other Internal Functions ---------------------------------------------------*/

/* deleteAssociation removes the association from the list of associations, frees all data allocated
   for it and <calls moduleprefix>_delete*(...) function at all modules.
   params: associationID: the Id of the association to be deleted.
*/
static void mdi_deleteAssociation(Association * assoc)
{
    if (assoc != NULL) {
        event_logi(INTERNAL_EVENT_0, "Deleting association %08x ", assoc->tagLocal);

        /* free module data */
        if (assoc->tagRemote != 0) {
            /* association init was allready completed */
            fc_delete_flowcontrol(assoc->flowControl);
            rtx_delete_reltransfer(assoc->reliableTransfer);
            rxc_delete_recvctrl(assoc->rx_control);
            se_delete_stream_engine(assoc->streamengine);
        }

        pm_deletePathman(assoc->pathMan);
        bu_delete(assoc->bundling);
        sci_deleteSCTP_control(assoc->sctp_control);
        releasePort(assoc->localPort);

        /* free association data */
        free(assoc->destinationAddresses);
        free(assoc);
    } else {
        error_log(ERROR_MAJOR, "mdi_deleteAssociation: association does not exist");
    }

    return;

}                               /* end: mdi_deleteAssociation */



/*------------------- Functions called by the Unix-Interface -------------------------------------*/
static void
mdi_receive_UDPmesg(Socket_descriptor socket_fd,
                    unsigned char *buffer,
                    int bufferLength, union sockunion *fromAddress, int fromAddressLength)
{
    guchar mesg_id;
    guchar mesg[3000];
    assoc_struct *as;
    gushort instance_name = ntohs(sctp_instance);
    ulp_send_struct *us;
    ulp_recv_struct *rs;
    guint32 assoc_id, result;
    short *idx = NULL;

    /* first byte is to contain the message id */
    mesg_id = (guchar) buffer[0];

    /* Java/ULP Client sends everything in NBO ! So do conversion to HBO first ! */
    /* then access the respective fields */

    event_logiii(VERBOSE, "Got UDP message %u on socket %d, len %u ", mesg_id,
                 socket_fd, bufferLength);
    /* Decode Message : change DISTRIB state */
    switch (mesg_id) {
    case ULP_REGISTER:
        if (!sctpInstance)
            error_log(ERROR_FATAL,
                      "sctpInstance not yet created, but Registration Request received");
        /* store new port number of peer instance. */
        sctpInstance->udp_peer_port = ntohs(fromAddress->sin.sin_port);
        event_logi(VERBOSE, "Registered ULP Registration Request on Port %u",
                   sctpInstance->udp_peer_port);
        /* TODO : if it is not 0, there is an error/restart */
        mesg[0] = ULP_REGISTER_OK;
        event_log(VERBOSE, "Got ULP_REGISTER message from peer ulp instance");
        memcpy(&mesg[1], &instance_name, sizeof(gushort));
        adl_send_message(socket_fd, mesg, 1 + sizeof(gushort), fromAddress);

        break;
    case ULP_INITIALIZE:
        error_log(ERROR_MAJOR, "ULP Initialize not implemented");
        break;
    case ULP_ASSOCIATE:
        as = (assoc_struct *) & buffer[1];

        as->SCTP_InstanceName = ntohs(as->SCTP_InstanceName);
        as->noOfOutStreams = ntohs(as->noOfOutStreams);
        as->destinationPort = ntohs(as->destinationPort);

        event_logii(VERBOSE,
                    "Associate : Instance Name : %u, OutStreams: %u, ",
                    as->SCTP_InstanceName, as->noOfOutStreams);
        event_logii(VERBOSE, "dest Port : %u, useLocalPort : %s",
                    as->destinationPort, (as->useLocalPort == TRUE) ? "TRUE" : "FALSE");

        assoc_id =
            sctp_associate(as->SCTP_InstanceName, as->noOfOutStreams,
                           (union sockunion *) &(as->destinationAddress),
                           as->destinationPort, as->useLocalPort);
        mesg[0] = ULP_ASSOCIATE_OK;
        assoc_id = htonl(assoc_id);
        memcpy(&(mesg[1]), &assoc_id, sizeof(guint32));
        adl_send_message(socket_fd, mesg, 1 + sizeof(guint32), fromAddress);
        break;
    case ULP_SHUTDOWN:
        memcpy(&assoc_id, &(buffer[1]), sizeof(guint32));
        assoc_id = ntohl(assoc_id);
        event_logii(VERBOSE, "ULP_SHUTDOWN : associationID: %u (%x)", assoc_id, assoc_id);

        result = sctp_shutdown(assoc_id);
        mesg[0] = ULP_SHUTDOWN_OK;
        memcpy(&(mesg[1]), &result, sizeof(guint32));
        adl_send_message(socket_fd, mesg, 1 + sizeof(guint32), fromAddress);
        break;
    case ULP_CLOSE:
        error_log(ERROR_MAJOR, "ULP Close not implemented");
        break;
    case ULP_SEND:
        us = (ulp_send_struct *) & buffer[1];
        us->associationID = ntohl(us->associationID);
        us->streamID = ntohs(us->streamID);
        us->length = ntohl(us->length);
        us->ppId = ntohl(us->ppId);
        us->destAddressIndex = ntohs(us->destAddressIndex);
        us->context = ntohl(us->context);
        event_logiii(VERBOSE,
                     "SCTP_SEND : associationID: %u, streamID: %u, length: %u",
                     us->associationID, us->streamID, us->length);
        event_logiii(VERBOSE,
                     "Protocol ID : %u, destAddressIndex: %d, context : %u",
                     us->ppId, us->destAddressIndex, us->context);

        if (us->length < 1500) {
            if (us->destAddressIndex == 0xFFFF) {
                result = sctp_sendChunk(us->associationID,
                                        &(us->streamID), (us->user_data),
                                        us->length, us->ppId, NULL,
                                        &(us->context), NULL, NULL, NULL);
            } else {
                result =
                    sctp_sendChunk(us->associationID, &(us->streamID),
                                   (us->user_data), us->length, us->ppId,
                                   &(us->destAddressIndex), &(us->context), NULL, NULL, NULL);
            }

        } else
            error_log(ERROR_FATAL, "Buffer to send is way too large !!!!!!!!");

        mesg[0] = ULP_SEND_OK;
        result = ntohl(result);
        memcpy(&(mesg[1]), &result, sizeof(guint32));
        adl_send_message(socket_fd, mesg, 1 + sizeof(guint32), fromAddress);
        break;

    case ULP_SET_PRIMARY:
        error_log(ERROR_MAJOR, "ULP Set Primary not implemented");
        break;
    case ULP_RECEIVE:
        rs = (ulp_recv_struct *) & buffer[1];
        rs->associationID = ntohl(rs->associationID);
        rs->streamID = ntohs(rs->streamID);
        rs->length = ntohl(rs->length);
        event_logiii(VERBOSE,
                     "ULP_UDP_RECEIVE : associationID: %u, streamID: %u, length: %u",
                     rs->associationID, rs->streamID, rs->length);
        result =
            sctp_receive(rs->associationID, rs->streamID,
                         &(mesg[1 + sizeof(guint32)]), &(rs->length));
        mesg[0] = ULP_RECEIVE_OK;
        memcpy(&(mesg[1]), &result, sizeof(guint32));
        adl_send_message(socket_fd, mesg, (1 + sizeof(guint32) + rs->length), fromAddress);
        break;
    case ULP_STATUS:
        error_log(ERROR_MAJOR, "ULP Status not implemented");
        break;
    case ULP_CHANGE_HB:
        error_log(ERROR_MAJOR, "ULP Change HB not implemented");
        break;
    case ULP_REQ_HB:
        error_log(ERROR_MAJOR, "ULP  Req HB not implemented");
        break;
    case ULP_GET_SRTT:
        error_log(ERROR_MAJOR, "ULP GET SRTT not implemented");
        break;

    }

}


/* mdi_receiveMessage is the callback function of the SCTP-message distribution. 
   It is called by the Unix-interface module when a new datagramm is received.
   @param socket_fd :         the socket file discriptor
   @param buffer :            DG
   @param bufferlength :      length of datagramm
   @param fromAddress :       source address of DG
   @param fromAddressLength : source address length
*/

/* CHECKME : Do we do correct OOTB handling ?
             Also, see e.g. ABORT-Handling with either local-tag OR peer-tag (8.5.1.B) */

void
mdi_receiveMessage(Socket_descriptor socket_fd,
                   unsigned char *buffer,
                   int bufferLength, union sockunion *fromAddress, int fromAddressLength)
{
    SCTP_message *message;
    SCTP_simple_chunk *chunk = NULL;
    SCTP_init_fixed *initChunk = NULL;
    guchar* initPtr;
    unsigned int i, len, state;
    boolean sourceAddressExists;
    boolean sendAbort = FALSE;

    boolean initFound = FALSE, cookieEchoFound = FALSE;

    guchar hoststring[MAX_MTU_SIZE];
    short shutdownCompleteCID;
    short abortCID;


    lastFromAddress = fromAddress;

    lastFromPath = 0;

    adl_sockunion2str(fromAddress, hoststring, MAX_MTU_SIZE);

    message = (SCTP_message *) buffer;

    if (!validate_datagram(buffer, bufferLength)) {
        event_log(INTERNAL_EVENT_0, "received corrupted datagramm");
        lastFromAddress = NULL;
        return;
    }

    len = bufferLength - sizeof(SCTP_common_header);

    /* save from address for response if a remote address is not available otherwise.
       For instance initAck or cookieAck. */
    lastFromPort = ntohs(message->common_header.src_port);
    lastDestPort = ntohs(message->common_header.dest_port);

    event_logiiii(EXTERNAL_EVENT,
                  "mdi_receiveMessage : len %d, sourceaddress : %s, src_port %u,dest_port %u",
                  bufferLength, hoststring, lastFromPort, lastDestPort);

/*  something we have to modify : check incoming (dest) port....this is for servers.
    Clients may chose their source port, and thus receive incoming datagrams to that
    port (sctpInstance->localPort is only significant for servers).
    Maybe we should add a client/server semantics to sctpInstance ?!

    if  (sctpInstance->localPort != 0) {
        if (lastDestPort != sctpInstance->localPort) {
            error_logii(ERROR_MINOR, "DG with dest port = %u received, sctpInstance->port = %u",
                  lastDestPort, sctpInstance->localPort);
            lastFromAddress = NULL;
            lastFromPort = 0;
            lastDestPort = 0;
            return;
    }    */

    if (lastFromPort == 0 || lastDestPort == 0) {
        error_log(ERROR_MINOR, "received DG with invalid (i.e. 0) ports");
        lastFromAddress = NULL;
        lastFromPort = 0;
        lastDestPort = 0;
        return;
    }

    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationByTransportAddress(lastFromAddress, lastFromPort);

    lastInitiateTag = ntohl(message->common_header.verification_tag);

    chunk = (SCTP_simple_chunk *) & message->sctp_pdu[0];

    /* check if sctp-message belongs to an existing association */
    if (currentAssociation == NULL) {
        event_log(VVERBOSE, "mdi_receiveMsg: currentAssociation==NULL, start scanning !");
        /* This is not very elegant, but....only used when assoc is being build up, so :-D */
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_ABORT) == TRUE) {
            event_log(INTERNAL_EVENT_0, "mdi_receiveMsg: Found ABORT chunk, discarding it !");
            lastFromAddress = NULL;
            lastFromPort = 0;
            lastDestPort = 0;
            return;
        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_SHUTDOWN_ACK) == TRUE) {
            event_log(INTERNAL_EVENT_0,
                      "mdi_receiveMsg: Found SHUTDOWN_ACK chunk, send SHUTDOWN_COMPLETE !");
            /* section 8.4.5 : return SHUTDOWN_COMPLETE with peers veri-tag and T-Bit set */
            shutdownCompleteCID = ch_makeSimpleChunk(CHUNK_SHUTDOWN_COMPLETE, FLAG_NO_TCB);
            bu_put_Ctrl_Chunk(ch_chunkString(shutdownCompleteCID));
            /* should send it to last address */
            bu_sendAllChunks(NULL);
            /* free abort chunk */
            ch_deleteChunk(shutdownCompleteCID);

            /* send an ABORT with peers veri-tag, set T-Bit */
            event_log(VERBOSE, "mdi_receiveMsg: sending CHUNK_SHUTDOWN_COMPLETE  ");
            lastFromPort = 0;
            lastDestPort = 0;
            return;
        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_SHUTDOWN_COMPLETE) == TRUE) {
            event_log(INTERNAL_EVENT_0,
                      "mdi_receiveMsg: Found SHUTDOWN_COMPLETE chunk, discarding it !");
            lastFromPort = 0;
            lastDestPort = 0;
            lastFromAddress = NULL;
            return;
        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_COOKIE_ACK) == TRUE) {
            event_log(INTERNAL_EVENT_0, "mdi_receiveMsg: Found COOKIE_ACK chunk, discarding it !");
            lastFromPort = 0;
            lastDestPort = 0;
            lastFromAddress = NULL;
            return;
        }

        /* section 8.4.7) : Discard the datagram, if it contains a STALE-COOKIE ERROR */
        if (bu_scanDatagramForError(message->sctp_pdu, len, ECC_STALE_COOKIE_ERROR) == TRUE) {
            event_log(INTERNAL_EVENT_0,
                      "mdi_receiveMsg: Found STALE COOKIE ERROR, discarding packet !");
            lastFromPort = 0;
            lastDestPort = 0;
            lastFromAddress = NULL;
            return;
        }
        if ((initPtr = bu_findChunk(message->sctp_pdu, len, CHUNK_INIT)) != NULL) {
            if (lastDestPort != sctpInstance->localPort || sctpInstance->localPort == 0) {
                /* destination port is not the listening port of this this SCTP-instance. */
                event_log(INTERNAL_EVENT_0,
                          "mdi_receiveMsg: INIT Message ignored, dest. port does not fit");
                sendAbort = TRUE;
                /* as per section 5.1 :
                   If an endpoint receives an INIT, INIT ACK, or COOKIE ECHO chunk but
                   decides not to establish the new association due to missing mandatory
                   parameters in the received INIT or INIT ACK, invalid parameter values,
                   or lack of local resources, it MUST respond with an ABORT chunk */
            } else {
                event_log(INTERNAL_EVENT_0, "mdi_receiveMsg: INIT Message - processing it !");

                /* TODO : check for all src address parameters in INIT, whether they correspond
                   to an association. If there are addresses previously unknown (that may have
                   been added or changed) discard the INIT. Might be an attack then.
                 */

            }
            initChunk = ((SCTP_init_fixed *) & ((SCTP_init *) message->sctp_pdu)->init_fixed);
            lastInitiateTag = ntohl(initChunk->init_tag);
            event_logi(VERBOSE, "setting lastInitiateTag to %x ", lastInitiateTag);

        } else if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_COOKIE_ECHO) == TRUE) {

            if (lastDestPort != sctpInstance->localPort || sctpInstance->localPort == 0) {
                /* destination port is not the listening port of this this SCTP-instance. */
                event_log(INTERNAL_EVENT_0,
                          "mdi_receiveMsg: COOKIE_ECHO ignored, dest. port does not fit");
                sendAbort = TRUE;
            } else {
                event_log(INTERNAL_EVENT_0,
                          "mdi_receiveMsg: COOKIE_ECHO Message - processing it !");
            }

        } else {
            /* section 8.4.8) send an ABORT with peers veri-tag, set T-Bit */
            event_log(INTERNAL_EVENT_0,
                      "mdi_receiveMsg: send ABORT -> message ignored (OOTB - see section 8.4.8) ");
            sendAbort = TRUE;
        }

        if (sendAbort == TRUE) {
            /* make and send abort message */
            abortCID = ch_makeSimpleChunk(CHUNK_ABORT, FLAG_NO_TCB);
            bu_put_Ctrl_Chunk(ch_chunkString(abortCID));
            /* should send it to last address */
            bu_sendAllChunks(NULL);
            /* free abort chunk */
            ch_deleteChunk(abortCID);
            /* send an ABORT with peers veri-tag, set T-Bit */
            event_log(VERBOSE, "mdi_receiveMsg: sending ABORT with T-Bit");
            lastFromAddress = NULL;
            lastFromPort = 0;
            lastDestPort = 0;
            /* and discard that packet */
            return;
        }

    } else { /* i.e. if(currentAssociation != NULL) */

        /* If the association exists, both ports of the message must be equal to the ports
           of the association and the source address must be in the addresslist of the peer
           of this association */
        /* check src- and dest-port and source address */
        if (lastFromPort != currentAssociation->remotePort
            || lastDestPort != currentAssociation->localPort) {
            error_logiiii(ERROR_MINOR,
                          "port mismatch in received DG (lastFromPort=%u, assoc->remotePort=%u, lastDestPort=%u, assoc->localPort=%u ",
                          lastFromPort, currentAssociation->remotePort,
                          lastDestPort, currentAssociation->localPort);
            currentAssociation = NULL;
            lastFromAddress = NULL;
            lastFromPort = 0;
            lastDestPort = 0;
            return;
        }

        /* check if source address is in address list of this association.
           tbd: check the draft if this is correct. */
        for (i = 0, sourceAddressExists = FALSE; i < currentAssociation->noOfNetworks; i++) {
            if (adl_equal_address
                (&(currentAssociation->destinationAddresses[i]), lastFromAddress) == TRUE) {
                sourceAddressExists = TRUE;
                break;
            }
        }

        if (!sourceAddressExists) {
            error_log(ERROR_MINOR,
                      "source address of received DG is not in the destination addresslist");
            currentAssociation = NULL;
            lastFromPort = 0;
            lastDestPort = 0;
            lastFromAddress = NULL;
            return;
        }

        if (sourceAddressExists)
            lastFromPath = i;

        /* check for verification tag rules --> see section 8.5 */
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_INIT) == TRUE) {
            /* check that there is ONLY init */
            initFound = TRUE;
            if (lastInitiateTag != 0) {
                currentAssociation = NULL;
                lastFromPort = 0;
                lastDestPort = 0;
                lastFromAddress = NULL;
                event_log(VERBOSE, "mdi_receiveMsg: scan found INIT, lastInitiateTag!=0, returning");
                return;
            }
            initChunk = ((SCTP_init_fixed *) & ((SCTP_init *) message->sctp_pdu)->init_fixed);
            /* make sure, if you send an ABORT later on (i.e. when peer requests 0 streams),
             * you pick the right tag */
            lastInitiateTag = ntohl(initChunk->init_tag);
            event_logi(VVERBOSE, "Got an INIT CHUNK with initiation-tag %u", lastInitiateTag);

        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_ABORT) == TRUE) {
            /* accept my-tag or peers tag, else drop packet */
            if ((lastInitiateTag != currentAssociation->tagLocal &&
                 lastInitiateTag != currentAssociation->tagRemote) || initFound == TRUE) {
                currentAssociation = NULL;
                lastFromPort = 0;
                lastDestPort = 0;
                lastFromAddress = NULL;
                return;
            }
        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_SHUTDOWN_COMPLETE) == TRUE) {
            /* accept my-tag or peers tag, else drop packet */
            /* TODO : make sure that if it is the peer's tag also T-Bit is set */
            if ((lastInitiateTag != currentAssociation->tagLocal &&
                 lastInitiateTag != currentAssociation->tagRemote) || initFound == TRUE) {
                currentAssociation = NULL;
                lastFromPort = 0;
                lastDestPort = 0;
                lastFromAddress = NULL;
                return;
            }
        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_SHUTDOWN_ACK) == TRUE) {
            if (initFound == TRUE) {
                currentAssociation = NULL;
                lastFromPort = 0;
                lastDestPort = 0;
                lastFromAddress = NULL;
                return;
            }
            state = sci_getState();
            if (state == COOKIE_ECHOED || state == COOKIE_WAIT) {
                /* see also section 8.5.E.) treat this like OOTB packet */
                event_logi(EXTERNAL_EVENT,
                           "mdi_receive_message: shutdownAck in state %u, send SHUTDOWN_COMPLETE ! ",
                           state);
                shutdownCompleteCID = ch_makeSimpleChunk(CHUNK_SHUTDOWN_COMPLETE, FLAG_NO_TCB);
                bu_put_Ctrl_Chunk(ch_chunkString(shutdownCompleteCID));
                bu_sendAllChunks(NULL);
                ch_deleteChunk(shutdownCompleteCID);
                currentAssociation = NULL;
                lastFromPort = 0;
                lastDestPort = 0;
                lastFromAddress = NULL;
                return;
            }
        }
        if (bu_scanDatagram(message->sctp_pdu, len, CHUNK_COOKIE_ECHO) == TRUE) {
               cookieEchoFound = TRUE;
        }

        if (!cookieEchoFound && !initFound && lastInitiateTag != currentAssociation->tagLocal) {
            error_logii(ERROR_MINOR,
                        "Tag mismatch in receive DG, received Tag = %u, local Tag = %u -> discarding",
                        lastInitiateTag, currentAssociation->tagLocal);

            currentAssociation = NULL;
            lastFromPort = 0;
            lastDestPort = 0;
            lastFromAddress = NULL;
            return;

        }

    }


    /* printString("Processing received SCTP-message : ", buffer, bufferLength); */

    /* forward DG to bundling */
    bu_rcvDatagram(lastFromPath, message->sctp_pdu, bufferLength - sizeof(SCTP_common_header));

    lastInitiateTag = 0;
    currentAssociation = NULL;
    lastFromAddress = NULL;
    lastFromPath = -1;          /* only valid for functions called via mdi_receiveMessage */

}                               /* end: mdi_receiveMessage */




/*------------------- Functions called by the ULP ------------------------------------------------*/
/*------------------- Prototypes are defined in sctp.h -------------------------------------------*/

/* sctp_registerSCTP_instance is called to initialize one SCTP-instance.
   Each Adaption-Layer of the ULP must create its own SCTP-instance.
   Each SCTP-instance has its own port.
   @param ulp_via_udp            TRUE is ulp and SCTP use defined UDP to communicate
   @param port                   wellknown port of sctp-instance
   @param noOfLocalAddresses     # of local addresses
   @param localAddressList       local address list
   @param ULPcallbackFunctions   call back functions for primitives past from sctp to ULP
   @return instance name of this SCTP-instance or -1 in case of errors
*/
unsigned short
sctp_registerSCTP_instance(gboolean ulp_via_udp,
                           unsigned short port,
                           unsigned int noOfLocalAddresses,
                           union sockunion *localAddressList,
                           struct ULPcallback ULPcallbackFunctions)
{

    DLL_Return exitCode;
    int adl_rscb_code;
    int i;
    struct timeval curTime;

    event_log(EXTERNAL_EVENT, "sctp_registerSCTP_instance called");

    if (firstSCTP_instance) {
        adl_init_adaptation_layer();

        socket_descriptor = adl_get_sctpv4_socket(port);
        if (!socket_descriptor)
            error_log(ERROR_FATAL, "socket creation failed");
        /* register callback functions for this sctp-instance */
#ifdef HAVE_IPV6
        if (support_ipv6) {
            ipv6_socket_descriptor = adl_get_sctpv6_socket(port);
            if (!ipv6_socket_descriptor)
                error_log(ERROR_FATAL, "IPv6 socket creation failed");
        } else {
            ipv6_socket_descriptor = 0;
        }
#endif

        adl_rscb_code = adl_register_socket_cb(socket_descriptor,
                                               &mdi_receiveMessage);
        if (!adl_rscb_code)
            error_log(ERROR_FATAL, "register socket call back function at adaptation failed");
#ifdef HAVE_IPV6
        /*
         * here some operating system specialties may kick in (i.e. opening only ONE
         * socket MIGHT be enough, provided IPv6 socket implicitly reveives IPv4 packets, too
         */
        if (support_ipv6) {
            adl_rscb_code = adl_register_socket_cb(ipv6_socket_descriptor,
                                                   &mdi_receiveMessage);
            if (!adl_rscb_code)
                error_log(ERROR_FATAL,
                          "register ipv6 socket call back function at adaptation failed");

        }
#endif
        if (ulp_via_udp == TRUE) {

            udp_socket_descriptor = adl_get_udp_socket(port);
            if (!udp_socket_descriptor)
                error_log(ERROR_FATAL, "UDP socket creation failed");

            adl_rscb_code = adl_register_socket_cb(udp_socket_descriptor,
                                                   &mdi_receive_UDPmesg);

            event_logi(VERBOSE, "register socket %d for UDP callbacks", udp_socket_descriptor);

        }

        /* Create list for associations */
        DLL_CreateList(&associations);
        if (associations == NULL)
            error_log(ERROR_FATAL, "List for associations could not be created");

        if ((exitCode = DLL_InitializeList(associations, sizeof(Association *))) != DLL_NORMAL)
            DLL_error_log(ERROR_FATAL, exitCode);

        /* initialize ports seized -- see comments above !!! */
        portsSeized = (unsigned char *) malloc(0x10000);
        for (i = 0; i < 0x10000; i++)
            portsSeized[i] = 0;
        lastPortSeized = 0x00000400;
        numberOfSeizedPorts = 0x00000000;

        /* initialize random number generator */
        gettimeofday(&curTime, NULL);
        rstate[0] = curTime.tv_sec;
        rstate[1] = curTime.tv_usec;
        initstate(curTime.tv_sec, (char *) rstate, 8);
        setstate((char *) rstate);

        /* initialize bundling, i.e. the common buffer for sending chunks when no association
           exists. */
        bu_init_bundling();

        /* this block is executed only once for the lifetime of sctp-software */
        firstSCTP_instance = FALSE;

        /* FOR BAKEOFF PURPOSES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
        /* 0 is the standard input ! */
        adl_rscb_code = adl_register_socket_cb(0, &ulp_stdin_cb);

        /* FOR BAKEOFF PURPOSES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
    }



    sctpInstance = (SCTP_instance *) malloc(sizeof(SCTP_instance));
    if (!sctpInstance)
        error_log_sys(ERROR_FATAL, errno);

    sctpInstance->sctpInstanceName = lastSCTP_instanceName++;
    sctpInstance->localPort = port;

    if (localAddressList != NULL) {
        sctpInstance->localAddressList =
            (union sockunion *) malloc(noOfLocalAddresses * sizeof(union sockunion));
        memcpy(sctpInstance->localAddressList, localAddressList,
               noOfLocalAddresses * sizeof(union sockunion));

        sctpInstance->noOfLocalAddresses = noOfLocalAddresses;
    }

    else {
        sctpInstance->noOfLocalAddresses = 0;
    }

    sctpInstance->ulp_via_udp = ulp_via_udp; /* i.e. handle all via udp, or all via timer cb functions */
    sctpInstance->udp_peer_port = 0; /* i.e. not yet set.....that is done with registration. */

    sctpInstance->ULPcallbackFunctions = ULPcallbackFunctions;

    return sctpInstance->sctpInstanceName;

}                               /* end: sctp_registerSCTP_instance */



/* This function is called to setup an association.
   The ULP must specify the SCTP-instance to which this association belongs to.
   @param SCTP_InstanceName :         the SCTP instance this association beleongs to.
   @param noOfOutStreams :            # of output stream the ULP would like to have
   @param destinationAddress :        destination address
   @param destinationPort :           destination port 
   @useLocalPort :                    if true, no new local port is assigned for the association.
                                      Instead the "listening" port of the sctp-instance is used
                                      as source port.
   returns: association ID of this association (identical with local tag).
            0 in case of failures
   remark: this does not conform to the current standard (draft version 07 from March 2, 2000,
           because there is useLocalPort 
*/
unsigned int
sctp_associate(unsigned short SCTP_InstanceName,
               unsigned short noOfOutStreams,
               union sockunion *destinationAddress,
               unsigned short destinationPort, boolean useLocalPort)
{
    unsigned int assocID;
    unsigned short zlocalPort;

    if (currentAssociation != NULL) {
        error_log(ERROR_MINOR, "sctp_associate: association not cleared");
        return 0;
    }

    event_log(EXTERNAL_EVENT, "sctp_associate called");

    if (useLocalPort)
        zlocalPort = sctpInstance->localPort;
    else
        zlocalPort = seizePort();

    /* Create new association */
    if (mdi_newAssociation(SCTP_InstanceName, zlocalPort, /* local client port */
                           destinationPort, /* remote server port */
                           mdi_generateLocalTag(), 0, 1, destinationAddress)) {
        error_log(ERROR_MAJOR, "Creation of association failed");
        return 0;
    }

    /* call associate at SCTP-control */
    scu_associate(noOfOutStreams, noOfOutStreams);

    assocID = currentAssociation->tagLocal;

    /* rest currentAssociation */
    currentAssociation = NULL;

    return assocID;
}                               /* end: sctp_associate */


/* sctp_shutdown initiates the shutdown of the specified association.
   @param    associationID : the ID of the addressed association.
   @return   error code
*/
int sctp_shutdown(unsigned int associationID)
{
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        /* Forward shutdown to the addressed association */
        scu_shutdown();
    } else {
        error_log(ERROR_MAJOR, "sctp_shutdown: addressed association does not exist");
        return 1;
    }

    /* clear association */
    if (clearAssociation)
        currentAssociation = NULL;

    return 0;

}                               /* end: sctp_shutdown */



/* sctp_abort initiates the abort of the specified association.
   @param    associationID : the ID of the addressed association.
   @return   error code
*/
int sctp_abort(unsigned int associationID)
{
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        /* Forward shutdown to the addressed association */
        scu_abort();
    } else {
        error_log(ERROR_MAJOR, "sctp_abort: addressed association does not exist");
        return 1;
    }

    /* clear association */
    if (clearAssociation)
        currentAssociation = NULL;

    return 0;

}                               /* end: sctp_abort */



/* sctp_sendChunk used by the ULP to send data chunks.
   @param    associationID :    the ID of the addressed association.
   @param    streamID :         identifies the stream on which the chunk is sent.
   @param    buffer             chunk data.
   @param    length             length of chunk data.
   @param    protocolId         the payload protocol identifier
   @param    destAddressIndex   index of destination address, if different from primary path.
   @param    context            ULP context for this data chunk, to be returned in case of errors.
   @param    lifetime           maximum time of chunk in send queue.
   @param    unorderedDelivery  the chunk is delivered to peer ULP without reseuencing.
   @param    dontBundle         chunk must not be bundled with other data chunks.
   @return   error code         -1 for send error, 1 for association error, 0 if successful
*/
int sctp_sendChunk(unsigned int associationID, unsigned short *streamID, /* optional (=null if none -> stream 0) */
                   unsigned char *buffer, unsigned int length, unsigned int protocolId, short *destAddressIndex, /* optional (=null if none) -> primary p. */
                   unsigned int *context, /* optional (=null if none) */
                   unsigned int *lifetime, /* optional (=null if none) -> infinit */
                   boolean * unorderedDelivery, /* optional (=null if none) -> ordered */
                   boolean * dontBundle)
{                               /* optional (=null if none) -> bundle */
    int result = 0;
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        if (destAddressIndex != NULL) {
            if (*destAddressIndex < 0 || *destAddressIndex >= currentAssociation->noOfNetworks) {
                error_log(ERROR_MAJOR, "sctp_sendChunk: invalid destination address");
                return 1;
            }
        }

        event_log(INTERNAL_EVENT_1, "sctp_send_chunk: sending chunk");
        /* Forward chunk to the addressed association */
        result = se_ulpsend(streamID, buffer, length, protocolId, destAddressIndex,
                      context, lifetime, unorderedDelivery, dontBundle);

    } else {
        error_log(ERROR_MAJOR, "sctp_sendChunk: addressed association does not exist");
        return 1;
    }

    /* clear association */
    if (clearAssociation)
        currentAssociation = NULL;

    return result;
}                               /* end: sctp_sendChunk */



/* sctp_setPrimary changes the primary path of an association.
   @param  associationID :    ID of assocation.
   @param  destAddressIndex : new primary path
   @return error code
*/
short sctp_setPrimary(unsigned int associationID, short destAddressIndex)
{
    short rv;
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        /* Forward shutdown to the addressed association */
        rv = pm_setPrimaryPath(destAddressIndex);
    } else {
        error_log(ERROR_MAJOR, "sctp_setPrimary: addressed association does not exist");
        return 1;
    }

    /* clear association */
    if (clearAssociation)
        currentAssociation = NULL;

    return rv;

}                               /* end: sctp_setPrimary */




/* sctp_receive is called in response to the dataArriveNotification to
   get the received data.
   The stream engine must copy the chunk data from a received  SCTP datagram to
   a new byte string, because the SCTP datagram is overwritten when the next datagram
   is received and the lifetime of a chunk in the streamengine might outlast the
   the reception of several SCTP datagrams.
   For this reasons and to avoid repeated copying of byte strings, a pointer to
   the byte string of chunkdata allocated by the streamengine is returned.
   According to the standard, the chunkdata should be copied to to a buffer provided
   by the ULP.
   @param   associationID : ID of association.
   @param   streamID :      the stream on which the data chunk is received.
   @param   buffer :        chunk data.
   @param   length :        length of chunk data.
   returns: one if association soes not exist, zero if okay.
*/
unsigned short
sctp_receive(unsigned int associationID,
             unsigned short streamID, unsigned char *buffer, unsigned int *length)
{
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        /* retreive data from streamengine instance */
        se_ulpreceive(buffer, length, streamID);
    } else {
        error_log(ERROR_MAJOR, "sctp_receive: addressed association does not exist");
        return 1;
    }

    /* clear association */
    if (clearAssociation)
        currentAssociation = NULL;

    return 0;
}                               /* end: sctp_receive */



/* associationStatus returns the status of an association.
   @param  associationID :    ID of assocation.
   @return association status or NULL if association does not exist.
*/
AssociationStatus *sctp_status(unsigned int associationID)
{
    AssociationStatus *tmp = NULL;
    guint16 num, result, count;

    event_logi(VERBOSE, "sctp_status: Association %u", associationID);

/*   struct associationStatus
{
    guint32                 association_state;
    guint16                 number_of_addresses;
    union sockunion         primaryDestinationAddress;
    gushort                 primaryAddressIndex;
    union sockunion         *destinationAddresses;

    gshort                  *path_states;
    guint32                 *SRTT;
    guint32                 *RTO;

    guint32                 currentReceiverWindowSize;
    guint32                 *currentCongestionWindowSize;
    guint32                 noOfChunksInSendQueue;

    guint32                 noOfChunksInRetransmissionQueue;
    guint32                 noOfChunksInReceptionQueue;
}; */

    /* Retrieve association from list  */
    result = mdi_setAssociationData(associationID);

    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL && result == 0) {
        event_logi(VERBOSE, "sctp_status: Association %u", associationID);

        tmp = malloc(sizeof(AssociationStatus));
        num = currentAssociation->noOfNetworks;
        tmp->destinationAddresses = malloc(num * sizeof(union sockunion));
        tmp->path_states = malloc(num * sizeof(gshort));
        tmp->SRTT = malloc(num * sizeof(guint32));
        tmp->RTO = malloc(num * sizeof(guint32));
        tmp->currentCongestionWindowSize = malloc(num * sizeof(guint32));

        tmp->association_state = sci_getState();

        tmp->number_of_addresses = num;
        memcpy(tmp->destinationAddresses,
               currentAssociation->destinationAddresses, num * sizeof(union sockunion));
        tmp->primaryAddressIndex = pm_readPrimaryPath();
        tmp->primaryDestinationAddress =
            currentAssociation->destinationAddresses[tmp->primaryAddressIndex];
        for (count = 0; count < num; count++) {
            tmp->path_states[count] = pm_readState(count);
            tmp->SRTT[count] = pm_readSRTT(count);
            tmp->RTO[count] = pm_readRTO(count);
        }

        tmp->currentReceiverWindowSize = rtx_read_remote_receiver_window();
        tmp->noOfChunksInRetransmissionQueue = rtx_readNumberOfUnackedChunks();

        fc_readCWNDs(tmp->currentCongestionWindowSize, num);

        tmp->noOfChunksInSendQueue = fc_readNumberOfQueuedChunks();

        tmp->noOfChunksInReceptionQueue = se_numOfQueuedChunks();

    } else {

        error_logi(ERROR_MAJOR, "sctp_status : association %u does not exist", associationID);

    }
    result = mdi_clearAssociationData(associationID);
    return tmp;

}                               /* end: sctp_status */




/* sctp_changeHeartBeat changes changes the hearbeat intervall of an association.
   @param  associationID :    ID of assocation.
   @param  heartbeatON :      heartbeat on or off
   @param  timeIntervall :    heartbeat time intervall in milliseconds
   @return error code
*/
int
sctp_changeHeartBeat(unsigned int associationID,
                     short destAddressIndex, boolean heartbeatON, unsigned int timeIntervall)
{
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        /* Forward change HB to the addressed association */
        if (heartbeatON) {
            pm_enableHB(destAddressIndex, timeIntervall);
            event_logii(VERBOSE,
                        "Setting HB interval for address %d to %u msecs !",
                        destAddressIndex, timeIntervall);
        } else
            pm_disableHB(destAddressIndex);
    } else {
        error_log(ERROR_MAJOR, "sctp_changeHeartBeat: addressed association does not exist");
        return 1;
    }

    /* clear association */
    if (clearAssociation)
        currentAssociation = NULL;

    return 0;

}                               /* end: sctp_changeHeartBeat */



/* sctp_requestHeartbeat initiates a heartbeat to the given address of an association.
   @param  associationID :      ID of assocation.
   @param  destAddressIndex :   destination address to which the heartbeat is sent.
   @return error code
*/
int sctp_requestHeartbeat(unsigned int associationID, short destAddressIndex)
{
    /* Retrieve association from list  */
    currentAssociation = retrieveAssociationULP(associationID);

    if (currentAssociation != NULL) {
        pm_doHB(destAddressIndex);
        event_logi(VERBOSE, "Sending HB on user request to path ID: %u !",
                        destAddressIndex);
    } else {
        error_log(ERROR_MAJOR, "sctp_requestHeartbeat: addressed association does not exist");
        return 1;
    }
    /* clear association */
    if (clearAssociation) currentAssociation = NULL;

    return 0;
}                               /* sctp_requestHeartbeat */



/* sctp_setFailureThreshold sets the threshold for retransmissions on the given address of an 
   association. If the threshold is exeeded, the the destination address is considered as 
   unreachable.
   @param  associationID :            ID of assocation.
   @param  destAddressIndex :         destination address that gets a new failure threshold.
   @param  pathMaxRetransmissions :   threshold for retransmissions.
   @return 0
*/
int
sctp_setFailureThreshold(unsigned int associationID,
                         short destAddressIndex, unsigned short pathMaxRetransmissions)
{
    error_log(ERROR_MAJOR, "sctp_setFailureThreshold: unimplemented funktion");
    return 0;
}                               /* end:  sctp_setFailureThreshold */



/* sctp_setProtocolParams sets protocol parameters on the given address of an association. 
   @param  associationID :            ID of assocation.
   @param  destAddressIndex :         destination address that gets a new failure threshold.
   @param  RTO_initial :              initial round trip time.
   @param  validCookieLife :          cookie life time.
   @param  assocMaxRetransmits :      max. # of retransmission before the peer is considered
                                      unreachable.
   @param  maxInitRetransmits :       max. # of retransmission before a destination address is
                                      considered as unreachable.
   @param  maxInitRetransmits :       max. # of retransmission the init and cookie messages.
   @return SRTT
*/
int sctp_setProtocolParams(unsigned int associationID, short destAddressIndex, /* optional (=null if none) */
                           unsigned int RTO_initial,
                           unsigned int validCookieLife, unsigned int assocMaxRetransmits, unsigned int pathMaxRetransmits, /* per path */
                           unsigned int maxInitRetransmits)
{
    error_log(ERROR_MAJOR, "sctp_setProtocolParams: unimplemented funktion");
    return 0;

}                               /* end: sctp_setProtocolParams */



/*------------------- Functions called by the SCTP bundling --------------------------------------*/

/* Used by bundling to send a SCTP-datagramm. 
    Bundling passes a static pointer and leaves space for common header, so:
        -> we can fill that header in up front !
   Before calling send_message at the adaption-layer, this function does:
   - add the SCTP common header to the message
   - convert the SCTP message to a byte string
   - retrieve the socket-file descriptor of the SCTP-instance
   - retrieve the destination address
    - retrieve destination port ???
   @param SCTP_message     SCTP message as a struct
   @param length           length of SCTP message.
   @param destAddresIndex  Index of address in the destination address list.
   @return                 Errorcode.
*/

int mdi_send_message(SCTP_message * message, unsigned int length, short destAddressIndex)
{
    union sockunion dest_su;
    SCTP_simple_chunk *chunk;

    int txmit_len = 0;
    guchar hoststring[MAX_MTU_SIZE];


    if (message == NULL) {
        error_log(ERROR_MINOR, "mdi_send_message: no message to send !!!");
        return 1;
    }

    chunk = (SCTP_simple_chunk *) & message->sctp_pdu[0];

    if (currentAssociation == NULL) {
        /* only possible case : initAck, no association exists yet,
           use last from address as destination address */

        if (lastFromAddress == NULL) {
            error_log(ERROR_MAJOR, "mdi_send_message: lastFromAddress does not exist for initAck");
            return 1;
        } else {
            /* only if the sctp-message received before contained an init-chunk */
            memcpy(&dest_su, lastFromAddress, sizeof(union sockunion));

            message->common_header.verification_tag = htonl(lastInitiateTag);
            /* write invalid tag value to lastInitiateTag (reset it) */
            lastInitiateTag = 0;
            /* swap ports */
            message->common_header.src_port = htons(mdi_readLastDestPort());
            message->common_header.dest_port = htons(mdi_readLastFromPort());
            event_logiii(VVERBOSE,
                         "mdi_send_message (I) : tag = %x, src_port = %u , dest_port = %u",
                         lastInitiateTag, mdi_readLastDestPort(), mdi_readLastFromPort());

        }
    } else {
        if (destAddressIndex < 0 || destAddressIndex >= currentAssociation->noOfNetworks) {
            error_log(ERROR_MAJOR, "mdi_send_message: invalid destination address");
            return 1;
        }

        /* Use given destination address from current association */
        memcpy(&dest_su,
               &(currentAssociation->destinationAddresses[destAddressIndex]),
               sizeof(union sockunion));


        if (isInitAckChunk(chunk)) {
            /* is true in case of an init-collision, normally when an initAck is sent
               no association exist and the last lastInitiateTag is used in the initAck. This
               is handled in the case above, where no association exists.
               Or when we respond to SHUTDOWN_ACK, see section 8.4.5)
             */
            if (lastInitiateTag == 0) {
                error_log(ERROR_MAJOR, "mdi_send_message: No verification tag");
                return 1;
            }

            message->common_header.verification_tag = htonl(lastInitiateTag);
        } else {
            message->common_header.verification_tag = htonl(currentAssociation->tagRemote);
        }

        message->common_header.src_port = htons(currentAssociation->localPort);
        message->common_header.dest_port = htons(currentAssociation->remotePort);
        event_logiii(VVERBOSE,
                     "mdi_send_message (II): tag = %x, src_port = %u , dest_port = %u",
                     ntohl(message->common_header.verification_tag),
                     currentAssociation->localPort, currentAssociation->remotePort);
    }

    /* calculate and insert checksum */
    validate_insert_checksum((unsigned char *) message, length);

    switch (sockunion_family(&dest_su)) {
    case AF_INET:
        txmit_len = adl_send_message(socket_descriptor, message, length, &dest_su);
        break;
#ifdef HAVE_IPV6
    case AF_INET6:
        if (support_ipv6) {
            txmit_len = adl_send_message(ipv6_socket_descriptor, message, length, &dest_su);
        } else {
            error_log(ERROR_MAJOR, "mdi_send_message: ipv6 not supported");
        }
        break;
#endif
    default:
        error_log(ERROR_MAJOR, "mdi_send_message: Unsupported AF_TYPE");
        break;
    }

    adl_sockunion2str(&dest_su, hoststring, MAX_MTU_SIZE);

    event_logiii(EXTERNAL_EVENT,
                 "sent SCTP message of %d bytes to %s, result was %d", length,
                 hoststring, txmit_len);

    return (txmit_len == length) ? 0 : -1;

}                               /* end: mdi_send_message */



/*------------------- Functions called by the SCTP to forward primitives to ULP ------------------*/


/*  indicates new data have arrived from peer (chapter 9.2.A).
    params: streamID 
*/
void mdi_dataArriveNotif(unsigned int streamID, unsigned int length, unsigned int protoID)
{
    if (currentAssociation != NULL) {
        /* tbd: retreive sctp-instance from list */
        /* Forward dataArriveNotif to the ULP */
        sctpInstance->ULPcallbackFunctions.dataArriveNotif(currentAssociation->tagLocal,
                                                           streamID, length, protoID);
    } else {
        error_log(ERROR_MAJOR, "mdi_dataArriveNotif: association not set");
    }
}                               /* end: mdi_dataArriveNotif */



/* indicates a change of network status (chapter 9.2.C).
   params: 1.  destinationAddresses
           2.  newState 
*/
void mdi_networkStatusChangeNotif(short destinationAddresses, unsigned short newState)
{
    if (currentAssociation != NULL) {
        sctpInstance->ULPcallbackFunctions.networkStatusChangeNotif(currentAssociation->tagLocal,
                                                                    destinationAddresses, newState);
    } else {
        error_log(ERROR_MAJOR, "mdi_networkStatusChangeNotif: association not set");
    }
}                               /* end: mdi_networkStatusChangeNotif */



/* indicates a send failure (chapter 9.2.B).
   params: pointer to data not sent 
           dataLength
           context from sendChunk 
*/
void mdi_sentFailureNotif(unsigned char *data, unsigned int dataLength, unsigned int *context)
{
    if (currentAssociation != NULL) {
        sctpInstance->ULPcallbackFunctions.sentFailureNotif(currentAssociation->tagLocal,
                                                            data, dataLength, context);
    } else {
        error_log(ERROR_MAJOR, "mdi_sentFailureNotif: association not set");
    }
}                               /* end: mdi_sentFailureNotif */



/* indicates that communication was lost to peer (chapter 9.2.E).
   params: status, type of event
*/
void mdi_communicationLostNotif(unsigned short status)
{
    if (currentAssociation != NULL) {
        sctpInstance->ULPcallbackFunctions.communicationLostNotif(currentAssociation->tagLocal,
                                                                  status);
    } else {
        error_log(ERROR_MAJOR, "mdi_communicationLostNotif: association not set");
    }
}                               /* end: mdi_communicationLostNotif */


int sctp_sendUlp(unsigned char mesgId, void *data)
{
    guchar mesg[3000];
    comm_up_struct *cus;
    comm_lost_struct *cls;
    nwStatChange *nw;
    dataArriveStruct *da;
    struct sockaddr_in destAddress;
    int socket_fd;

    if (!sctpInstance) {
        error_log(ERROR_MAJOR, "SCTP Instance not set yet, exiting mdi_ulpNotif()");
        return -1;
    }
    if (sctpInstance->udp_peer_port == 0) {
        error_log(ERROR_MAJOR, "ULP Peer Port not set yet, exiting mdi_ulpNotif()");
        return -2;
    }
    destAddress.sin_family = AF_INET;
    destAddress.sin_port = htons(sctpInstance->udp_peer_port);
    destAddress.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    socket_fd = adl_get_udp_socket(sctpInstance->udp_peer_port);

    switch (mesgId) {
    case ULP_COMM_UP:
        mesg[0] = ULP_COMM_UP;
        event_log(VERBOSE, "Sending ULP_COMM_UP message to peer ulp instance");
        cus = (comm_up_struct *) data;
        cus->assoc_id = htonl(cus->assoc_id);
        cus->status = htons(cus->status);
        cus->noOfDestinations = htonl(cus->noOfDestinations);
        cus->noOfInStreams = htons(cus->noOfInStreams);
        cus->noOfOutStreams = htons(cus->noOfOutStreams);
        memcpy(&mesg[1], cus, sizeof(comm_up_struct));
        adl_send_message(socket_fd, mesg, 1 + sizeof(comm_up_struct),
                         (union sockunion *) &destAddress);
        break;
    case ULP_DATA_ARRIVE:
        mesg[0] = ULP_DATA_ARRIVE;
        da = (dataArriveStruct *) data;
        da->assoc_id = htonl(da->assoc_id);
        da->stream_id = htonl(da->stream_id);
        da->len = htonl(da->len);
        memcpy(&mesg[1], da, sizeof(dataArriveStruct));
        event_log(VERBOSE, "Sending ULP_DATA_ARRIVE message to peer ulp instance");
        adl_send_message(socket_fd, mesg, 1 + sizeof(dataArriveStruct),
                         (union sockunion *) &destAddress);
        break;
    case ULP_SEND_FAILURE:
        error_log(ERROR_MAJOR, "ULP_SEND_FAILURE not yet implemented ");
        break;
    case ULP_NETWORK_STATUS_CHANGE:
        mesg[0] = ULP_NETWORK_STATUS_CHANGE;
        nw = (nwStatChange *) data;
        nw->assoc_id = htonl(nw->assoc_id);
        nw->dest_add = htons(nw->dest_add);
        nw->new_state = htons(nw->new_state);
        memcpy(&mesg[1], nw, sizeof(nwStatChange));
        adl_send_message(socket_fd, mesg, 1 + sizeof(nwStatChange),
                         (union sockunion *) &destAddress);
        break;
    case ULP_COMM_LOST:
        mesg[0] = ULP_COMM_LOST;
        cls = (comm_lost_struct *) data;
        cls->assoc_id = htonl(cls->assoc_id);
        cls->status = htons(cls->status);
        cls->chunks_received = htonl(cls->chunks_received);
        memcpy(&mesg[1], cls, sizeof(comm_lost_struct));
        adl_send_message(socket_fd, mesg, 1 + sizeof(comm_lost_struct),
                         (union sockunion *) &destAddress);
        break;
    case ULP_COMM_ERR:
    case ULP_RESTART_NOTIFICATION:
    case ULP_SHUTDOWN_COMPLETE:
    default:
        error_logi(ERROR_MAJOR, "Message %u not yet implemented ", mesgId);
    }
    return 0;
}

/* indicates that a association is established (chapter 9.2.D).
   params: status, type of event
*/
void mdi_communicationUpNotif(unsigned short status)
{
    union sockunion *lastAddress;
    short primaryPath;
    unsigned short noOfInStreams;
    unsigned short noOfOutStreams;

    if (currentAssociation != NULL) {
        /* Find primary path */
        lastAddress = (union sockunion *) mdi_readLastFromAddress();

        if (lastAddress != NULL) {

            for (primaryPath = 0; primaryPath < currentAssociation->noOfNetworks; primaryPath++) {
                if (adl_equal_address
                    (&(currentAssociation->destinationAddresses[primaryPath]), lastAddress)) {
                    break;
                }
            }
        } else {
            primaryPath = 0;
        }
        if (primaryPath >= currentAssociation->noOfNetworks)
            primaryPath = 0;

        /* set number of paths and primary path at pathmanegement and start heartbeat */
        pm_setPaths(currentAssociation->noOfNetworks, primaryPath);

        se_readNumberOfStreams(&noOfInStreams, &noOfOutStreams);


        event_logiii(VERBOSE,
                     "Distribution: COMM-UP, assocId: %u, status: %u, noOfNetworks: %u",
                     currentAssociation->tagLocal, status, currentAssociation->noOfNetworks);
        event_logii(VERBOSE, "noOfInStreams: %u,noOfOutStreams  %u", noOfInStreams, noOfOutStreams);
        /* tbd: retreive sctp-instance from list */
        /* Forward mdi_communicationup Notification to the ULP */
        sctpInstance->ULPcallbackFunctions.communicationUpNotif(currentAssociation->tagLocal,
                                                                status,
                                                                currentAssociation->
                                                                destinationAddresses,
                                                                currentAssociation->noOfNetworks,
                                                                noOfInStreams, noOfOutStreams);

        free(lastAddress);
    } else {
        error_log(ERROR_MAJOR, "mdi_communicationUpNotif: association not set");
    }
}                               /* end: mdi_communicationLostNotif */



/*------------------- Functions called by the SCTP to get current association data----------------*/

/* When processing external events from outside the SCTP (socket events, timer events and 
   function calls from the ULP), first the data of the addressed association are read
   from the list of associations and stored in a private but static datastructure.
   Elements of this association data can be read by the following functions.
*/


/* The following functions return pointer to data of modules of the SCTP. As only these
   modules know the exact type of these data structures, so the returned pointer are
   of type void.
*/

/*
 * returns: pointer to the flow control data structure.
 *          null in case of error.
 */
void *mdi_readFlowControl()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readFlowControl: association not set");
        return NULL;
    } else {
        return currentAssociation->flowControl;
    }
}



/*
 * returns: pointer to the reliable transfer data structure.
 *          null in case of error.
 */
void *mdi_readReliableTransfer()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readReliableTransfer: association not set");
        return NULL;
    } else {
        return currentAssociation->reliableTransfer;
    }
}



/*
 * returns: pointer to the RX-control data structure.
 *          null in case of error.
 */
void *mdi_readRX_control()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readRX_control: association not set");
        return NULL;
    } else {
        return currentAssociation->rx_control;
    }
}



/*
 * returns: pointer to the stream engine data structure.
 *          null in case of error.
 */
void *mdi_readStreamEngine()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readStreamEngine: association not set");
        return NULL;
    } else {
        return currentAssociation->streamengine;
    }
}



/*
 * returns: pointer to the pathmanagement data structure.
 *          null in case of error.
 */
void *mdi_readPathMan()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readPathMan: association not set");
        return NULL;
    } else {
        return currentAssociation->pathMan;
    }
}


/*
 * returns: pointer to the bundling data structure.
 *          null in case of error.
 */
void *mdi_readBundling()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readBundling: association not set");
        return NULL;
    } else {
        return currentAssociation->bundling;
    }
}



/*
 * returns: pointer to the SCTP-control data structure.
 *          null in case of error.
 */
void *mdi_readSCTP_control()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readSCTP_control: association not set");
        return NULL;
    } else {
        return currentAssociation->sctp_control;
    }
}



/* 
 * returns: association-ID of the curent association
 *          0 means the association is not set (an error).
 */
unsigned int mdi_readAssociationID()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readAssociationID: association not set");
        return 0;
    } else {
        return currentAssociation->tagLocal;
    }
}



unsigned int mdi_readTagRemote()
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_readAssociationID: association not set");
        return 0;
    } else {
        return currentAssociation->tagRemote;
    }
}



unsigned int mdi_generateLocalTag()
{
    unsigned int tag;

    while ((tag = random()) == 0);

    return tag;
}



unsigned int mdi_generateStartTSN()
{
    return random();
}



/*------------- functions to for the cookie mechanism --------------------------------------------*/

/* each time a datagram is received, its source address, source port and destination port
   is saved. This address-info is used in the following cases:
   - z-side: for sending the initAck (no association exists from which the addresses could be read).
   - a-side: if the initAck  ??
   - z-side: if the cookie-chunk does not contain addresses of the a-side, lastFromAddress_in
             is used as the only destination address.
*/


/* returns the address from which the last datagramm was received (host byte order).
*/
union sockunion *mdi_readLastFromAddress()
{
    union sockunion *destAddressList;

    if (lastFromAddress == NULL) {
        error_log(ERROR_MINOR, "mdi_saveLastFromAddress: no last from address");
        return NULL;
    } else {
        destAddressList = (union sockunion *) malloc(sizeof(union sockunion));
        if (destAddressList != NULL) {
            memcpy(destAddressList, lastFromAddress, sizeof(union sockunion));
        } else {
            error_log(ERROR_MAJOR, "memory allocation failed ! Aborting ");
            return NULL;
        }
        return destAddressList;
    }
}


/* reads the path from which the last DG was received. -1 is returned if no DG was received.
*/
short mdi_readLastFromPath()
{
    return lastFromPath;
}

/* returns the port of the sender of the last received DG (host byte order).
*/
unsigned short mdi_readLastFromPort()
{
    if (lastFromAddress == NULL) {
        error_log(ERROR_MINOR, "readLastFromPort: no last from address");
        return 0;
    } else {
        return lastFromPort;
    }
}


/* returns the port of the destination of the last received DG (host byte order).
*/
unsigned short mdi_readLastDestPort()
{
    if (lastFromAddress == NULL) {
        error_log(ERROR_MINOR, "readLastDestPort: no last from address");

        return 0;
    } else {
        return lastDestPort;
    }
}

/* write the initiate tag of a-side to be used as verification tag for the initAck */
void mdi_writeLastInitiateTag(unsigned int initiateTag)
{
    lastInitiateTag = initiateTag;
}

/* rewrite the initiate tag of peer in case of a peer reset. */
void mdi_rewriteTagRemote(unsigned int newInitiateTag)
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_rewriteRemoteTag: association not set");
    } else {
        currentAssociation->tagRemote = newInitiateTag;
    }
}

/*------------- functions to write and read addresses --------------------------------------------*/

void mdi_writeDestinationAddresses(union sockunion addresses[20], int noOfAddresses)
{

    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_writeDestinationAddresses: association not set");
        return;
    } else {
        if (currentAssociation->destinationAddresses != NULL)
            free(currentAssociation->destinationAddresses);

        currentAssociation->destinationAddresses =
            (union sockunion *) malloc(noOfAddresses * sizeof(union sockunion));

        if (currentAssociation->destinationAddresses == NULL)
            error_log(ERROR_FATAL, "mdi_writeDestinationAddresses: out of memory");

        memcpy(currentAssociation->destinationAddresses, addresses,
               noOfAddresses * sizeof(union sockunion));

        currentAssociation->noOfNetworks = noOfAddresses;

        return;
    }
}



void mdi_readLocalAddresses(union sockunion addresses[20], guint16 * noOfAddresses)
{

    if (currentAssociation == NULL && mdi_readLastDestPort() == 0) {
        error_log(ERROR_MINOR, "mdi_readLocalAddresses: association not set");
        *noOfAddresses = 0;
        return;
    }
    if (currentAssociation == NULL) {
        /* tbd: retrieve SCTP-instance with last destination port */
        ;
    } else {
        /* tbd: retrieve SCTP-instance with SCTP-instance name in current association */
        ;
    }

    event_logi(VERBOSE, "mdi_readLocalAddresses() : returning %u addresses !",
               sctpInstance->noOfLocalAddresses);

    memcpy(addresses, sctpInstance->localAddressList,
           sctpInstance->noOfLocalAddresses * sizeof(union sockunion));

    *noOfAddresses = sctpInstance->noOfLocalAddresses;
}



/*------------- functions to set and clear the association data ----------------------------------*/

/* Each module within SCTP that has timers implements its own timer call back
   functions. These are registered at the adaption layer when a timer is started
   and called directly at the module when the timer expires.
   setAssociationData allows SCTP-modules with timers to retrieve the data of the
   addressed association from the list of associations.
   For this purpose the association-ID must be included in one of the 
   parameters of the start_timer function of the adaption-layer.
   Params:  associationID: the ID of the association
   returns: 0 if successful
            1 if the association does not exist in the list.
*/
unsigned short mdi_setAssociationData(unsigned int associationID)
{
    if (currentAssociation != NULL)
        error_log(ERROR_MINOR, "mdi_setAssociationData: previous assoc not cleared");

    /* retrieve association from list */
    currentAssociation = retrieveAssociation(associationID);
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_setAssociationData: association does not exist");
        return 1;
    }
    return 0;
}



/* Clear the global association data.
   This function must be called after the association retrieved from the list
   with setAssociationData is no longer needed. This is the case after a time
   event has been handled.
   returns:  0 if successfull.
             1 if association data have not been set.
             2 wrong associationID
*/
unsigned short mdi_clearAssociationData(unsigned int associationID)
{
    if (currentAssociation == NULL) {
        error_log(ERROR_MINOR, "mdi_clearAssociationData: association not set");
        return 1;
    }

    if (currentAssociation->tagLocal != associationID) {
        error_log(ERROR_MINOR, "mdi_clearAssociationData: wrong associationID");
        return 2;
    }

    /* clear currentAssociation */
    currentAssociation = NULL;

    return 0;
}


/*------------------- Functions to create and delete associations --------------------------------*/

/* 
   This function allocates memory for a new association.
   For the active side of an association, this function is called when the associate is called from 
   the ULP.
   For the passive side this function is called when a cookie message is received.
   It also creates all the modules pathman, bundling and SCTP-control.
   The rest of the modules are created with mdi_initAssociation.
   The created association is entered into the list of associations.
   If successful, a pointer to the newly created association structure is return.
   NULL in case of failures.
*/
unsigned short
mdi_newAssociation(unsigned short SCTP_InstanceName,
                   unsigned short local_port,
                   unsigned short remote_port,
                   unsigned int tagLocal,
                   short primaryDestinitionAddress,
                   short noOfDestinationAddresses, union sockunion *destinationAddressList)
{

    /* tbd: retrieve SCTP-instance using local port from list */
    /* check if SCTP-instance exists */
    if (!sctpInstance) {
        error_log(ERROR_MAJOR, "SCTP-protocol not ready");
        return 1;
    }

    /* If no SCTP-instance name is provided (z-side), use the instance name
       from the SCTP-instance found in list. */
    if (SCTP_InstanceName == 0) {
        SCTP_InstanceName = sctpInstance->sctpInstanceName;
    }

    /* check wether SCTP-instance names are equal (makes sense only for the a-side) */
    if (sctpInstance->sctpInstanceName != SCTP_InstanceName) {
        error_log(ERROR_MAJOR,
                  "Tried to establish an association within a nonexistent SCTP-instance");
        return 1;
    }

    /* Do plausi checks on the addresses. */
    if (noOfDestinationAddresses <= 0 || destinationAddressList == NULL) {
        error_log(ERROR_MAJOR, "No destination address suppllied for new association");
        return 1;
    }
        else if (primaryDestinitionAddress < 0
                 || primaryDestinitionAddress >= noOfDestinationAddresses) {
        error_log(ERROR_MAJOR, "Invalid primary destination address for new association");
        return 1;
    }

    if (currentAssociation) {
        error_log(ERROR_MINOR, "current association not cleared");
    }

    currentAssociation = (Association *) malloc(sizeof(Association));

    if (!currentAssociation) {
        error_log_sys(ERROR_FATAL, errno);
        return 1;
    }

    currentAssociation->sctpInstanceName = SCTP_InstanceName;
    currentAssociation->localPort = local_port;
    currentAssociation->remotePort = remote_port;
    currentAssociation->tagLocal = tagLocal;
    currentAssociation->tagRemote = 0;
    currentAssociation->deleted = FALSE;

    currentAssociation->noOfNetworks = noOfDestinationAddresses;
    currentAssociation->destinationAddresses =
        (union sockunion *) malloc(noOfDestinationAddresses * sizeof(union sockunion));
    memcpy(currentAssociation->destinationAddresses, destinationAddressList,
           noOfDestinationAddresses * sizeof(union sockunion));

    /* initialize pointer to other modules of SCTP */
    currentAssociation->flowControl = NULL;
    currentAssociation->reliableTransfer = NULL;
    currentAssociation->rx_control = NULL;
    currentAssociation->streamengine = NULL;

    /* only pathman, bundling and sctp-control are created at this point, the rest is created
       with mdi_initAssociation */
    currentAssociation->bundling = bu_new();
    currentAssociation->pathMan = pm_newPathman(noOfDestinationAddresses,
                                                primaryDestinitionAddress);
    currentAssociation->sctp_control = sci_newSCTP_control();

    event_logi(INTERNAL_EVENT_1, "new Association created ID=%08x", currentAssociation->tagLocal);

    /* check if newly created association already exists. */
    if (checkForExistingAssociations(currentAssociation)) {
        error_log(ERROR_MAJOR, "tried to establish an existing association");
        /* FIXME : also free bundling, pathmanagement,sctp_control */
        free(currentAssociation);
        currentAssociation = NULL;
        return 1;
    }

    /* Enter association into list */
    enterAssociation(currentAssociation);

    return 0;

}                               /* end: mdi_newAssociation */


/* An association is created in two steps because data become available at the a-side in two steps:
   1. associate
   2. init acknowledgement
   At the z-side, with the cookie message all data are available at once. So mdi_newAssociation
   and mdi_initAssociation must be called when the initAck is recieved.
*/
unsigned short
mdi_initAssociation(unsigned int remoteSideReceiverWindow,
                    unsigned short noOfInStreams,
                    unsigned short noOfOutStreams,
                    unsigned int remoteInitialTSN,
                    unsigned int tagRemote, unsigned int localInitialTSN)
{
    if (!currentAssociation) {
        error_log(ERROR_MAJOR,
                  "mdi_initAssociation: current association does not exist, can not initialize");
        return 1;
    }

    /* if  mdi_initAssociation has already be called, delete modules and make new ones
       with possibly new data. Multiple calls of of mdi_initAssociation can occur on the
       a-side in the case of stale cookie errors. */
    if (currentAssociation->tagRemote != 0) {
        event_log(INTERNAL_EVENT_1,
                  "Deleting Modules in mdi_initAssociation() -- then recreating them !!!!");
        /* association init was already completed */
        fc_delete_flowcontrol(currentAssociation->flowControl);
        rtx_delete_reltransfer(currentAssociation->reliableTransfer);
        rxc_delete_recvctrl(currentAssociation->rx_control);
        se_delete_stream_engine(currentAssociation->streamengine);
    }

    /* TODO : check number of input and output streams (although that should be fixed now) */

    currentAssociation->tagRemote = tagRemote;
    currentAssociation->reliableTransfer =
        (void *) rtx_new_reltransfer(currentAssociation->noOfNetworks);
    currentAssociation->flowControl =
        (void *) fc_new_flowcontrol(remoteSideReceiverWindow, localInitialTSN,
                                    currentAssociation->noOfNetworks);

    currentAssociation->rx_control = (void *) rxc_new_recvctrl(remoteInitialTSN);
    currentAssociation->streamengine = (void *) se_new_stream_engine(noOfInStreams, noOfOutStreams);
    event_logi(INTERNAL_EVENT_1,
               "second step of association initialisation performed ID=%08x",
               currentAssociation->tagLocal);

    return 0;

}                               /* end: mdi_initAssociation */


/* mdi_deleteCurrentAssociation deletes the current association. The association will not be deleted
   at once, but is only marked for deletion. This is done in this way to allow other modules to finish
   there current activities. To prevent them to start new activities, the currentAssociation pointer
   is set to NULL. */
void mdi_deleteCurrentAssociation()
{
    short pathID;

    if (currentAssociation != NULL) {
        if (currentAssociation->tagRemote != 0) {
            /* stop timers */
            for (pathID = 0; pathID < currentAssociation->noOfNetworks; pathID++)
                pm_disableHB(pathID);

            fc_stop_timers();
            rxc_stop_sack_timer();
        }

        /* mark association as deleted, it will be deleted when retrieveAssociation(..) encounters
           a "deleted" association. */
        currentAssociation->deleted = TRUE;
        event_logi(INTERNAL_EVENT_1, "association ID=%08x marked for deletion",
                   currentAssociation->tagLocal);
        mdi_clearAssociationData(currentAssociation->tagLocal);
    } else {
        error_log(ERROR_MAJOR,
                  "mdi_initAssociation: current association does not exist, can not delete");
    }
}
