/*
 *  $Id: sctp.h,v 1.35 2000/10/31 17:51:38 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: This header-file defines the interface for the ULP, the application
 *          that uses SCTP.
 *          Further comments on message distribution can be found in the implemen-
 *          tation of this interface distribution.c.
 * 
 * function prefix: sctp_         
 */

#ifndef SCTP_H
#define SCTP_H


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



#include  "globals.h"           /* boolean, etc */


/******************** Defines *********************************************************************/

/* Pathstatus, used with networkstatus primitives */
#define PATH_OK              0
#define PATH_UNREACHABLE     1

/* **************** Definitions for communication with the ULP via a UDP socket ***************** */

/* For the initial handshake of the ULP instance and SCTP instance */
#define ULP_REGISTER            1
#define ULP_REGISTER_OK                  2

/* messages from SCTP to the ULP */
#define ULP_DATA_ARRIVE             7
typedef struct __dataArriveStruct
{
    unsigned int assoc_id;
    unsigned int stream_id;
    unsigned int len;
} dataArriveStruct;
#define ULP_SEND_FAILURE            8
#define ULP_NETWORK_STATUS_CHANGE   9
typedef struct __nw_stat_change
{
    unsigned int assoc_id;
    unsigned short dest_add;
    unsigned short new_state;
} nwStatChange;
#define ULP_COMM_UP                 10
typedef struct __comm_up_struct
{
    unsigned int assoc_id;
    unsigned int noOfDestinations;
    unsigned short status;
    unsigned short noOfInStreams;
    unsigned short noOfOutStreams;
} comm_up_struct;
#define ULP_COMM_LOST               11
typedef struct __comm_lost_struct
{
    unsigned int assoc_id;
    unsigned short status;
    unsigned int chunks_received;
} comm_lost_struct;

#define ULP_COMM_ERR                12
#define ULP_RESTART_NOTIFICATION    13
#define ULP_SHUTDOWN_COMPLETE       14

/* messages from SCTP to the ULP (for returning results) */
#define ULP_INITIALIZE_OK           15
#define ULP_ASSOCIATE_OK            16
#define ULP_SHUTDOWN_OK             17
#define ULP_CLOSE_OK                18
#define ULP_SEND_OK                 19
#define ULP_SET_PRIMARY_OK          20
#define ULP_RECEIVE_OK              21
#define ULP_STATUS_OK               22
#define ULP_CHANGE_HB_OK            23
#define ULP_REQ_HB_OK               24
#define ULP_GET_SRTT_OK             25

/* messages from the ULP to the SCTP instance */
#define ULP_INITIALIZE              28
#define ULP_ASSOCIATE               29
typedef struct __assoc_struct
{
    unsigned short SCTP_InstanceName;
    unsigned short noOfOutStreams;
    unsigned short destinationPort;
    unsigned short useLocalPort; /* 0 for true, 1 for false */
// #ifdef HAVE_IPV6
//    struct sockaddr_in6 destinationAddress;
//#else
    struct sockaddr_in destinationAddress;
//#endif
}
assoc_struct;

#define ULP_SHUTDOWN                30
#define ULP_CLOSE                   31
#define ULP_SEND                    32
typedef struct __ulp_send_struct
{
    unsigned int associationID;
    unsigned int length;
    unsigned int ppId;
    unsigned short streamID;
    unsigned short destAddressIndex; /* 0xFFFF for primary path, else the one specified */
    unsigned int context;       /* optional (=0 if none) */
    unsigned short unorderedDelivery; /* 0 if true, 1 if false */
    unsigned char user_data[1500];
} ulp_send_struct;

#define ULP_SET_PRIMARY             33
#define ULP_RECEIVE                 34
typedef struct __ulp_recv_struct
{
    unsigned int associationID;
    unsigned short streamID;
    unsigned int length;
    unsigned char user_data[1500];
} ulp_recv_struct;
#define ULP_STATUS                  135
#define ULP_CHANGE_HB               136
#define ULP_REQ_HB                  137
#define ULP_GET_SRTT                138
/* There are a few more which I do not need right now */

/******************** Structure Definitions *******************************************************/


/* This struct containes the pointers to ULP callback functions.
   Each SCTP-instance can have its own set of callback functions.
   The callback functions of each SCTP-instance can be found by
   first reading the datastruct of an association from the list of 
   associations. The datastruct of the association contains the name
   of the SCTP instance to which it belongs. With the name of the SCTP-
   instance its datastruct can be read from the list of SCTP-instances.
*/

struct ULPcallback
{
    /*  indicates new data have arrived from peer (chapter 9.2.A).
       params: 1. associationID
       2. streamID
       3. length of data
     */
    void (*dataArriveNotif) (unsigned int, unsigned int, unsigned int, unsigned int);

    /* indicates a change of network status (chapter 9.2.C).
       params: 1.  associationID
       2.  destinationAddresses
       3.  newState 
     */
    void (*networkStatusChangeNotif) (unsigned int, short, unsigned short);

    /* indicates a sned failure (chapter 9.2.B).
       params: 1.  associationID
       2.  pointer to data not sent 
       3.  dataLength
       4.  context from sendChunk 
     */
    void (*sentFailureNotif) (unsigned int, unsigned char *, unsigned int, unsigned int *);

    /* indicates that communication was lost to peer (chapter 9.2.E).
       params: 1.  associationID
       2.  status, type of event
     */
    void (*communicationLostNotif) (unsigned int, unsigned short);

    /* indicates that a association is established (chapter 9.2.D).
       params: 1.  associationID
       2.  status, type of event
       3.  destination addresses
       4.  number of destination addresses
       5.  number input streamns
       6.  number output streams
     */
    void (*communicationUpNotif) (unsigned int, unsigned short,
                                  union sockunion *, int, unsigned short, unsigned short);
};


/*
 *  This struct contains the data to be returned to the ULP with the
 *  status primitive. Pointers point to arrays indexed by the number of
 *  destination addresses.
 */
typedef struct SCTP_Status
{
    gushort association_state;
    guint16 number_of_addresses;
    union sockunion primaryDestinationAddress;
    gushort primaryAddressIndex;
    union sockunion *destinationAddresses; /*  from SCTP-Control                   */
    /*  from PathManagement                 */
    gshort *path_states;        /*  PM_ACTIVE     0, PM_INACTIVE   1    */
    guint32 *SRTT;              /*  im msecs                            */
    guint32 *RTO;               /*  im msecs                            */

    guint32 currentReceiverWindowSize; /* from FlowControl                     */
    guint32 *currentCongestionWindowSize;
    guint32 noOfChunksInSendQueue;

    guint32 noOfChunksInRetransmissionQueue; /* from ReliableTransfer                */
    guint32 noOfChunksInReceptionQueue; /* from StreamEngine ?                  */
}
AssociationStatus;



/******************** Function Definitions ********************************************************/

/* This function is called to initialize one SCTP-instance.
   Each Adaption-Layer of the ULP must create its own SCTP-instance.
   Each SCTP-instance has its own port.
   returns: instance name of this SCTP-instance or -1 in case of errors
*/
unsigned short sctp_registerSCTP_instance(gboolean ulp_via_udp,
                                          unsigned short localPort,
                                          unsigned int noOfLocalAddresses,
                                          union sockunion *localAddressList,
                                          struct ULPcallback ULPcallbackFunctions);

/* This function is called to setup an association.
   The ULP must specify the SCTP-instance to which this association belongs to.
   @param SCTP_InstanceName :         the SCTP instance this association beleongs to.
   @param noOfOutStreams :            # of output stream the ULP would like to have
   @param destinationAddress :        destination address
   @param destinationPort :           destination port 
   @useLocalPort :                    if true, no new local port is assigned for the association.
                                      Instead the "listening" port of the sctp-instance is used
                                      as source port.
   returns: association ID of this association (identical with local tag).
            0 in case of failures
   remark: this does not conform to the current standard (draft version 07 from March 2, 2000,
           because there is useLocalPort 
*/
unsigned int sctp_associate(unsigned short SCTP_InstanceName,
                            unsigned short noOfOutStreams,
                            union sockunion *destinationAddress,
                            unsigned short destinationPort, boolean useLocalPort);



/* initiates the shutdown of the specified association.
   returns: error code
*/
int sctp_shutdown(unsigned int associationID);



/* aborts the specified association.
   returns the errorcode
*/
int sctp_abort(unsigned int associationID);


/* ULP requests the transmission of a datachunk.     
   returns the errorcode
*/
int sctp_sendChunk(unsigned int associationID, unsigned short *streamID, /* optional (=null if none -> stream 0) */
                   unsigned char *buffer, unsigned int length, unsigned int protocolId, short *destAddressIndex, /* optional (=null if none) -> primary p. */
                   unsigned int *context, /* optional (=null if none) */
                   unsigned int *lifetime, /* optional (=null if none) -> infinit */
                   boolean * unorderedDelivery, /* optional (=null if none) -> ordered */
                   boolean * dontBundle); /* optional (=null if none) -> bundle */


/* returns the errorcode */
short sctp_setPrimary(unsigned int associationID, short destAddressIndex);




/* This function is called in response to the dataArriveNotification to
   get the received data.
   The stream engine must copy the chunk data from a received UDP-datagram to
   a new byte string, because the UDP-datagram overwritten when the next datagram
   is received and the lifetime of a chunk in the streamengine might outlast the
   the reception of several UDP-datagrams.
   For this reasons and to avoid repeated copying of byte strings, a pointer to
   the byte string of chunkdata allocated by the streamengine is returned.
   According to the standard, the chunkdata should be copied to to a buffer provided
   by the ULP.
   Params: In:      associationID:  ID of association
                    streamID:       ID of stream
           Out:     buffer:        the buffer containing the bytestring of the received DG.
                    length:        length of DG.
   returns: zero if nothing is available
*/
unsigned short sctp_receive(unsigned int associationID,
                            unsigned short streamID, unsigned char *buffer, unsigned int *length);



/* Used by the ULP to read the status of the specified association.
   returns: the status of the association,
            null if the association doesn't exist 
*/
AssociationStatus *sctp_status(unsigned int associationID);




/* Used to switch on or off the heartbeat or change the 
   timeinterval of the heartbeat.
   returns: error code
*/
int sctp_changeHeartBeat(unsigned int associationID,
                         short destAddressIndex, boolean heartbeatON, unsigned int timeIntervall);



/* initiates a heartbeat for the specified destination
   returns: error code
*/
int sctp_requestHeartbeat(unsigned int associationID, short destAddressIndex);



/* Reads the SRTT of the specified path.
   returns: SRTT
*/
unsigned int sctp_getSRTT_Report(unsigned int associationID, short destAddressIndex);



/* Sets the threshold for the number of retransmissions, after which
   the path is recognized as no longer available.
   returns: error code
*/
int sctp_setFailureThreshold(unsigned int associationID,
                             short destAddressIndex, unsigned short pathMaxRetransmissions);



/* This function is used to set/modify the protocol parameters.
   returns: error code
*/
int sctp_setProtocolParams(unsigned int associationID, short destAddressIndex, /* optional (=null if none) */
                           unsigned int RTO_initial,
                           unsigned int validCookieLife, unsigned int assocMaxRetransmits, unsigned int pathMaxRetransmits, /* per path */
                           unsigned int maxInitRetransmits);


/**
 * Function for notifying an associated ULP instance (after it has registered
 * with the SCTP instance) of a callback event (COMM_UP, COMM_LOST, DATA_ARRIVE etc.)
 * via local UDP socket
 */
int sctp_sendUlp(unsigned char mesgId, void *data);


/**
 * this function is to return the association id belonging to the destination address
 * Association must be ESTABLISHED !
 */
unsigned int sctp_get_assocID(union sockunion * destination);

extern int adl_event_loop();

#endif
